from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models
from .addressing import Addressing
from .cloud import Cloud
from .dhcp import Dhcp
from .fhrp import Fhrp
from .interfaces import Interfaces
from .ip_telephony import IpTelephony
from .load_balancing import LoadBalancing
from .managed_networks import ManagedNetworks
from .management import Management
from .mpls import Mpls
from .multicast import Multicast
from .neighbors import Neighbors
from .oam import Oam
from .platforms import Platforms
from .port_channels import PortChannels
from .qos import Qos
from .routing import Routing
from .sdn import Sdn
from .sdwan import Sdwan
from .security import Security
from .stp import Stp
from .vlans import Vlans
from .wireless import Wireless


class Technology(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def platforms(self):
        return Platforms(client=self.client, sn=self.sn)

    @property
    def interfaces(self):
        return Interfaces(client=self.client, sn=self.sn)

    @property
    def neighbors(self):
        return Neighbors(client=self.client, sn=self.sn)

    @property
    def dhcp(self):
        return Dhcp(client=self.client, sn=self.sn)

    @property
    def port_channels(self):
        return PortChannels(client=self.client, sn=self.sn)

    @property
    def vlans(self):
        return Vlans(client=self.client, sn=self.sn)

    @property
    def stp(self):
        return Stp(client=self.client, sn=self.sn)

    @property
    def addressing(self):
        return Addressing(client=self.client, sn=self.sn)

    @property
    def fhrp(self):
        return Fhrp(client=self.client, sn=self.sn)

    @property
    def managed_networks(self):
        return ManagedNetworks(client=self.client, sn=self.sn)

    @property
    def mpls(self):
        return Mpls(client=self.client, sn=self.sn)

    @property
    def multicast(self):
        return Multicast(client=self.client, sn=self.sn)

    @property
    def cloud(self):
        return Cloud(client=self.client, sn=self.sn)

    @property
    def management(self):
        return Management(client=self.client, sn=self.sn)

    @property
    def ip_telephony(self):
        return IpTelephony(client=self.client, sn=self.sn)

    @property
    def load_balancing(self):
        return LoadBalancing(client=self.client, sn=self.sn)

    @property
    def oam(self):
        return Oam(client=self.client, sn=self.sn)

    @property
    def qos(self):
        return Qos(client=self.client, sn=self.sn)

    @property
    def routing(self):
        return Routing(client=self.client, sn=self.sn)

    @property
    def sdn(self):
        return Sdn(client=self.client, sn=self.sn)

    @property
    def sdwan(self):
        return Sdwan(client=self.client, sn=self.sn)

    @property
    def security(self):
        return Security(client=self.client, sn=self.sn)

    @property
    def wireless(self):
        return Wireless(client=self.client, sn=self.sn)

    @property
    def serial_ports(self):
        return models.Table(client=self.client, endpoint="tables/serial-ports", sn=self.sn)


__all__ = ["Technology"]



import json
import logging
from collections import OrderedDict
from http.cookiejar import CookieJar
from importlib import metadata
from ipaddress import IPv4Address, AddressValueError
from typing import Optional, Union, Dict, List, Any
from urllib.parse import urljoin
from uuid import UUID
from warnings import warn

import dotenv
from httpx import Client, URL

from ipfabric.auth import Settings, AccessToken
from ipfabric.models import Snapshot, SNAPSHOT_COLUMNS, OAS
from ipfabric.settings.user_mgmt import User
from ipfabric.tools import VALID_REFS

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

OAS_DIR = files("ipfabric.oas")

logger = logging.getLogger("ipfabric")

LAST_ID, PREV_ID, LASTLOCKED_ID = VALID_REFS


class IPFabricAPI(Client):
    def __init__(
        self,
        base_url: Optional[str] = None,
        api_version: Optional[str] = None,
        snapshot_id: Optional[str] = None,
        auth: Optional[Any] = None,
        unloaded: bool = False,
        env_file: Optional[str] = None,
        streaming: bool = True,
        **kwargs,
    ):
        """Initializes the IP Fabric Client

        Args:
            base_url: IP Fabric instance provided in 'base_url' parameter, or the 'IPF_URL' environment variable
            api_version: [Optional] Version of IP Fabric API
            auth: API token, tuple (username, password), or custom Auth to pass to httpx
            snapshot_id: IP Fabric snapshot ID to use by default for database actions - defaults to '$last'
            unloaded: True to load metadata from unloaded snapshots
            env_file: Path to .env file to load
            **kwargs: Keyword args to pass to httpx
        """
        if {kwargs.get(i, None) for i in ["token", "username", "password"]} != {None}:  # TODO: Remove v6.8
            warn(
                "`token=''` or `username='', password=''` authentication will be deprecated in v6.8.0.",
                DeprecationWarning,
                stacklevel=2,
            )
            logger.warning(
                "Use of `token='<TOKEN>'` or `username='<USER>', password='<PASS>'` authentication will be deprecated "
                "in v6.8.0, please use `auth='<TOKEN>'` or `auth=('<USER>','<PASS>')` instead.\n"
                "This does not apply to .env file or environment variables (IPF_TOKEN, IPF_USERNAME, IPF_PASSWORD).\n"
                "This is to support custom authentication methods that will be passed directly to HTTPX."
            )
        self.unloaded = unloaded
        # find env file
        dotenv.load_dotenv(env_file if env_file else dotenv.find_dotenv(usecwd=True))
        with Settings() as settings:
            cookie_jar = CookieJar()
            super().__init__(
                cookies=cookie_jar,
                headers={"User-Agent": f'python-ipfabric-sdk/{metadata.version("ipfabric")}'},
                **self._httpx_kwargs(kwargs, kwargs.get("verify", settings.verify), settings.timeout),
            )
            base_url = base_url or settings.url
            if not base_url:
                raise RuntimeError("IP Fabric base_url not provided or IPF_URL not set")

            self.api_version, self.os_version = self.check_version(api_version or settings.version, base_url)
            self.base_url = urljoin(base_url, f"api/{self.api_version}/")
            snapshot_id = snapshot_id or settings.snapshot
            if auth:
                token, username, password = None, None, None
            else:
                token = kwargs.get("token", settings.token)  # TODO: Update this in v6.8
                username = kwargs.get("username", settings.username)
                password = kwargs.get("password", settings.password)

        if token:
            self._login(token)
        elif username and password:
            self._login((username, password), base_url=base_url, cookie_jar=cookie_jar)
        else:
            self._login(auth, base_url=base_url, cookie_jar=cookie_jar)  # TODO: Keep only this in v6.8

        # Get Current User, by doing that we are also ensuring the token is valid
        self._prev_snapshot_id = None
        self.user = self.get_user()
        self.snapshots = self.get_snapshots()
        self.oas = self._get_oas(base_url)
        self.streaming = streaming
        self._attribute_filters = None
        self._no_loaded_snapshots = False
        self.snapshot_id = snapshot_id
        logger.debug(
            f"Successfully connected to '{self.base_url.host}' IPF version '{self.os_version}' "
            f"as user '{self.user.username}'"
        )

    @property
    def _api_insuf_rights(self):
        msg = f'API_INSUFFICIENT_RIGHTS for user "{self.user.username}" '
        if self.user.token:
            msg += f'token "{self.user.token.description}" '
        return msg

    @property
    def hostname(self):
        resp = self.get("/os/hostname")
        if resp.status_code == 200:
            return resp.json()["hostname"]
        elif resp.status_code == 403:
            logger.critical(self._api_insuf_rights + 'on GET "/os/hostname".')
            return None
        else:
            resp.raise_for_status()

    @property
    def web_to_api(self) -> Dict[str, OAS]:
        return {v.web_endpoint: v for k, v in self.oas.items() if v.web_endpoint}

    def _get_oas(self, base_url: str) -> Dict[str, OAS]:
        try:
            oas = {k: OAS(**v) for k, v in json.loads(OAS_DIR.joinpath(self.api_version + ".json").read_text()).items()}
            return oas
        except FileNotFoundError:
            return self._parse_oas(base_url)

    def _parse_oas(self, base_url) -> Dict[str, OAS]:
        resp = self.get(urljoin(base_url, "/api/oas/openapi-extended.json"))
        resp.raise_for_status()

        endpoints = dict()
        for endpoint, methods in resp.json()["paths"].items():
            data = OAS(api_endpoint=endpoint[1:])
            if "post" not in methods:
                continue
            try:
                data.web_endpoint = methods["post"]["x-table"]["webPath"]
            except KeyError:
                pass
            try:
                columns = set(
                    methods["post"]["requestBody"]["content"]["application/json"]["schema"]["properties"]["columns"][
                        "items"
                    ]["enum"]
                )
                data.columns = list(columns)
            except KeyError:
                pass
            try:
                columns = methods["post"]["responses"]["200"]["content"]["application/json"]["schema"]["properties"][
                    "data"
                ]["items"]["properties"]
                data.nested_columns = [k for k, v in columns.items() if "type" in v and v["type"] == "array"]
            except KeyError:
                pass
            endpoints[endpoint[1:]] = data

        return endpoints

    @staticmethod
    def _httpx_kwargs(kwargs: dict, verify: Any, timeout: Optional[int]):
        httpx_kwargs = kwargs.copy()
        # TODO: Remove 'token', 'username', 'password' in v6.8
        [httpx_kwargs.pop(h, None) for h in ["cookies", "token", "username", "password"]]
        httpx_kwargs["verify"] = verify
        if "timeout" not in kwargs and timeout:
            httpx_kwargs["timeout"] = timeout
        return httpx_kwargs

    def _login(self, auth: Any, base_url: str = None, cookie_jar: CookieJar = None):
        if not auth:
            raise RuntimeError("IP Fabric Authentication not provided.")
        elif isinstance(auth, str):
            self.headers.update({"X-API-Token": auth})
        elif isinstance(auth, tuple):
            resp = self.post("auth/login", json=dict(username=auth[0], password=auth[1]))
            resp.raise_for_status()
            self.auth = AccessToken(Client(base_url=base_url, cookies=cookie_jar))
        else:
            self.auth = auth

    @property
    def attribute_filters(self):
        return self._attribute_filters

    @attribute_filters.setter
    def attribute_filters(self, attribute_filters: Union[Dict[str, List[str]], None]):
        if attribute_filters:
            logger.warning(
                "Setting Global Attribute Filter for all tables/diagrams until explicitly unset to None.\n"
                "Adding an Attribute Filter to any function will overwrite the Global Filter.\n"
                f"Filter: {attribute_filters}"
            )
        self._attribute_filters = attribute_filters

    def get_user(self) -> User:
        """Gets current logged in user information.

        Returns:
            User: User model of logged in user
        """
        resp = self.get("users/me")
        resp.raise_for_status()
        return User(**resp.json())

    def check_version(self, api_version: str = None, base_url: Union[URL, str] = None) -> tuple:
        """Checks API Version and returns the version to use in the URL and the OS Version

        Args:
            api_version: User defined API Version or None
            base_url: URL of IP Fabric

        Returns:
            api_version, os_version
        """
        api_version = (
            api_version.lstrip("v").split(".") if api_version else metadata.version("ipfabric").lstrip("v").split(".")
        )

        resp = self.get(URL(base_url or self.base_url).join("/api/version"))
        resp.raise_for_status()
        os_api_version = resp.json()["apiVersion"].lstrip("v").split(".")
        return_version = f"v{api_version[0]}.{api_version[1]}" if len(api_version) > 1 else f"v{api_version[0]}"
        if len(api_version) == 1 and api_version[0] > os_api_version[0]:
            logger.warning(
                f"Specified API or SDK Version (v{api_version[0]}) is greater then "
                f"OS API Version. Using OS Version:  (v{os_api_version[0]})"
            )
            return_version = f"v{os_api_version[0]}"
        elif api_version[0:2] > os_api_version[0:2]:
            logger.warning(
                f"Specified API or SDK Version (v{'.'.join(api_version)}) is greater then "
                f"OS API Version. Using OS Version:  (v{'.'.join(os_api_version)})"
            )
            return_version = f"v{os_api_version[0]}.{os_api_version[1]}"
        elif os_api_version[0] > api_version[0]:
            raise RuntimeError(
                f"OS Major Version v{os_api_version[0]} is greater then SDK Version "
                f"v{api_version[0]}.  Please upgrade the Python SDK to the new major version."
            )

        return return_version, resp.json()["releaseVersion"]

    def update(self):
        """get all snapshots and assigns them to an attribute"""
        self.snapshots = self.get_snapshots()
        self._no_loaded_snapshots = self.loaded_snapshots == dict()

    @property
    def loaded_snapshots(self) -> Dict[str, Snapshot]:
        """get only loaded snapshots"""
        return {k: v for k, v in self.snapshots.items() if v.loaded}

    @property
    def unloaded_snapshots(self) -> Dict[str, Snapshot]:
        if not self.unloaded:
            logger.warning("Unloaded snapshots not initialized. Retrieving unloaded snapshots.")
            self.unloaded = True
            self.update()
        return {k: v for k, v in self.snapshots.items() if not v.loaded}

    @property
    def snapshot_id(self) -> str:
        """get snapshot Id"""
        return self._snapshot_id

    @property
    def prev_snapshot_id(self) -> str:
        """get previous snapshot Id"""
        return self._prev_snapshot_id

    @property
    def snapshot(self) -> Snapshot:
        return self.snapshots[self.snapshot_id]

    @snapshot_id.setter
    def snapshot_id(self, snapshot_id):
        snapshot_id = snapshot_id or LAST_ID
        if not self.loaded_snapshots:
            logger.warning("No Snapshots are currently loaded.  Please load a snapshot before querying any data.")
            self._snapshot_id = None
            self._no_loaded_snapshots = True
        elif snapshot_id not in self.snapshots:
            # Verify snapshot ID is valid
            raise ValueError(f"Incorrect Snapshot ID: '{snapshot_id}'")
        else:
            self._prev_snapshot_id = getattr(self, "_snapshot_id", None)
            self._snapshot_id = self.snapshots[snapshot_id].snapshot_id

    def get_snapshot(self, snapshot_id: str):
        if snapshot_id in self.snapshots:
            return self.snapshots[snapshot_id]
        else:
            if self.api_version < "v6.3":  # TODO: Remove v7.0
                SNAPSHOT_COLUMNS.discard("creatorUsername")
            payload = {"columns": list(SNAPSHOT_COLUMNS), "filters": {"id": ["eq", snapshot_id]}}
            results = self._ipf_pager("tables/management/snapshots", payload)
            if not results:
                logger.error(f"Snapshot {snapshot_id} not found.")
                return None
            get_results = self._get_snapshots()
            snapshot = self._create_snapshot_model(results[0], get_results)
            if snapshot.loaded:
                snapshot.get_assurance_engine_settings(self)
            return snapshot

    def _create_snapshot_model(self, s, get_results):
        return Snapshot(
            client=self,
            **s,
            licensedDevCount=get_results[s["id"]].get("licensedDevCount", None),
            errors=get_results[s["id"]].get("errors", None),
            version=get_results[s["id"]]["version"],
            initialVersion=get_results[s["id"]].get("initialVersion", None),
        )

    def get_snapshot_id(self, snapshot: Union[Snapshot, str]):
        """
        Returns a Snapshot ID for a given input.

        Args:
            snapshot: Snapshot model, name, or ID

        Returns:
            Snapshot ID
        """
        if isinstance(snapshot, Snapshot):
            return snapshot.snapshot_id
        elif snapshot in VALID_REFS:
            return self.snapshots[snapshot].snapshot_id
        try:
            UUID(snapshot, version=4)
            return self.snapshots[snapshot].snapshot_id
        except ValueError:
            for snap in list(self.snapshots.values()):
                if snapshot == snap.name:
                    return snap.snapshot_id
        raise ValueError(f"Could not locate Snapshot ID for {snapshot}.")

    def _get_snapshots(self):
        """
        Need to do a GET and POST to get all Snapshot data. See NIM-7223
        POST Missing:
        licensedDevCount
        errors
        version
        initialVersion
        """
        res = self.get("/snapshots")
        res.raise_for_status()
        return {s["id"]: s for s in res.json()}

    def get_snapshots(self) -> Dict[str, Snapshot]:
        """Gets all snapshots from IP Fabric and returns a dictionary of {ID: Snapshot_info}

        Returns:
            Dictionary with ID as key and dictionary with info as the value
        """
        if self.api_version < "v6.3":  # TODO: Remove v7.0
            SNAPSHOT_COLUMNS.discard("creatorUsername")
        payload = {"columns": list(SNAPSHOT_COLUMNS), "sort": {"order": "desc", "column": "tsEnd"}}
        if not self.unloaded:
            logger.warning("Retrieving only loaded snapshots. To load all snapshots set `unloaded` to True.")
            payload["filters"] = {"and": [{"status": ["eq", "done"]}, {"finishStatus": ["eq", "done"]}]}
        results = self._ipf_pager("tables/management/snapshots", payload)
        get_results = self._get_snapshots()

        snap_dict = OrderedDict()
        for s in results:
            snap = self._create_snapshot_model(s, get_results)
            snap_dict[snap.snapshot_id] = snap
            if snap.loaded:
                snap.get_assurance_engine_settings()
                if LASTLOCKED_ID not in snap_dict and snap.locked:
                    snap_dict[LASTLOCKED_ID] = snap
                if LAST_ID not in snap_dict:
                    snap_dict[LAST_ID] = snap
                    continue
                if PREV_ID not in snap_dict:
                    snap_dict[PREV_ID] = snap
        return snap_dict

    def _ipf_pager(
        self,
        url: str,
        payload: dict,
        limit: int = 1000,
        start: int = 0,
    ):
        """
        Loops through and collects all the data from the tables
        :param url: str: Full URL to post to
        :param payload: dict: Data to submit to IP Fabric
        :param start: int: Where to start for the data
        :return: list: List of dictionaries
        """
        payload["pagination"] = dict(limit=limit)
        data = list()

        def page(s):
            payload["pagination"]["start"] = s
            r = self.post(url, json=payload)
            r.raise_for_status()
            return r.json()["data"]

        r_data = page(start)
        data.extend(r_data)
        while limit == len(r_data):
            start = start + limit
            r_data = page(start)
            data.extend(r_data)
        return data

    def trigger_backup(self, sn: str = None, ip: str = None):
        if sn:
            payload = {"sn": sn}
        else:
            try:
                ip = str(IPv4Address(ip))
            except AddressValueError:
                raise ValueError(f"Invalid IP Address, CIDR is not allowed: {ip}")
            payload = {"ip": ip}
        res = self.post("/discovery/trigger-config-backup", json=payload)
        res.raise_for_status()
        return res.status_code



from .device import Device, Devices
from .intent import Intent
from .inventory import Inventory
from .jobs import Jobs
from .oas import OAS
from .snapshot import Snapshot, SNAPSHOT_COLUMNS
from .table import BaseTable, Table
from .technology import Technology

__all__ = [
    "BaseTable",
    "Table",
    "Inventory",
    "Technology",
    "Jobs",
    "Snapshot",
    "Intent",
    "SNAPSHOT_COLUMNS",
    "Device",
    "Devices",
    "OAS",
]



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Qos(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def policy_maps(self):
        return models.Table(client=self.client, endpoint="tables/qos/policy-maps", sn=self.sn)

    @property
    def shaping(self):
        return models.Table(client=self.client, endpoint="tables/qos/shaping", sn=self.sn)

    @property
    def queuing(self):
        return models.Table(client=self.client, endpoint="tables/qos/queuing", sn=self.sn)

    @property
    def policing(self):
        return models.Table(client=self.client, endpoint="tables/qos/policing", sn=self.sn)

    @property
    def priority_queuing(self):
        return models.Table(client=self.client, endpoint="tables/qos/priority-queuing", sn=self.sn)

    @property
    def marking(self):
        return models.Table(client=self.client, endpoint="tables/qos/marking", sn=self.sn)

    @property
    def random_drops(self):
        return models.Table(client=self.client, endpoint="tables/qos/random-drops", sn=self.sn)



import logging
from time import sleep
from typing import Any, Literal

from pydantic import BaseModel

from .table import BaseTable

logger = logging.getLogger("ipfabric")

SNAP_JOBS = {
    "load": "snapshotLoad",
    "unload": "snapshotUnload",
    "download": "snapshotDownload",
    "add": "discoveryAdd",
    "refresh": "discoveryRefresh",
    "delete": "deleteDevice",
}

SNAP_ACTIONS = Literal["load", "unload", "download", "add", "refresh", "delete"]


class Jobs(BaseModel):
    client: Any = None

    @property
    def all_jobs(self):
        return BaseTable(client=self.client, endpoint="tables/jobs")

    @property
    def columns(self):
        return [
            "id",
            "downloadFile",
            "finishedAt",
            "isDone",
            "name",
            "scheduledAt",
            "snapshot",
            "startedAt",
            "status",
            "username",
        ]

    def _return_job_when_done(self, job_filter: dict, retry: int = 5, timeout: int = 5):
        """
        Returns the finished job. Only supports Snapshot related Jobs
        Args:
            job_filter: table filter for jobs
            retry: how many times to query the table
            timeout: how long to wait in-between retries

        Returns:
            job: list[dict[str, str]]: a job that has a status of done
        """
        if "name" not in job_filter and "snapshot" not in job_filter:
            raise SyntaxError("Must provide a Snapshot ID and name for a filter.")
        sleep(1)  # give the IPF server a chance to start the job
        # find the running snapshotDownload job (i.e. not done)
        jobs = self.all_jobs.fetch(
            filters=job_filter,
            sort={"order": "desc", "column": "startedAt"},
            columns=self.columns,
        )
        logger.debug(f"job filter: {job_filter}\nlist of jobs:{jobs}")
        if jobs:
            # if the job is already completed, we return it
            if jobs[0]["isDone"]:
                return jobs[0]
            # use the start time of this job to identify this specific job
            start_time = jobs[0]["startedAt"]
            job_filter["startedAt"] = ["eq", start_time]
            logger.debug(f"new job_filter: {job_filter}")

            for retries in range(retry):
                jobs = self.all_jobs.fetch(
                    filters=job_filter, sort={"order": "desc", "column": "startedAt"}, columns=self.columns
                )
                logger.debug(f"Current job: {jobs}")
                if jobs and jobs[0]["isDone"]:
                    return jobs[0]
                logger.info(
                    f"{job_filter['name'][1]} job is not ready for snapshot"
                    f" {job_filter['snapshot'][1]} ({retries}/{retry})"
                )
                sleep(timeout)
        else:
            logger.debug(f"job not found: {job_filter}")
        return None

    def check_snapshot_job(
        self, snapshot_id: str, started: int, action: SNAP_ACTIONS, retry: int = 5, timeout: int = 5
    ):
        """Checks to see if a snapshot load job is completed.

        Args:
            snapshot_id: UUID of a snapshot
            started: Integer time since epoch in milliseconds
            action: Type of job to filter on
            timeout: How long in seconds to wait before retry
            retry: how many retries to use when looking for a job, increase for large downloads

        Returns:
            Job dictionary if load is completed, None if still loading
        """
        j_filter = dict(snapshot=["eq", snapshot_id], name=["eq", SNAP_JOBS[action]], startedAt=["gte", started - 100])
        return self._return_job_when_done(j_filter, retry=retry, timeout=timeout)

    def check_snapshot_assurance_jobs(
        self, snapshot_id: str, assurance_settings: dict, started: int, retry: int = 5, timeout: int = 5
    ):
        """Checks to see if a snapshot Assurance Engine calculation jobs are completed.

        Args:
            snapshot_id: UUID of a snapshot
            assurance_settings: Dictionary from Snapshot.get_assurance_engine_settings
            started: Integer time since epoch in milliseconds
            timeout: How long in seconds to wait before retry
            retry: how many retries to use when looking for a job, increase for large downloads

        Returns:
            True if load is completed, False if still loading
        """
        j_filter = dict(snapshot=["eq", snapshot_id], name=["eq", "loadGraphCache"], startedAt=["gte", started - 100])
        if (
            assurance_settings["disabled_graph_cache"] is False
            and self._return_job_when_done(j_filter, retry=retry, timeout=timeout) is None
        ):
            logger.error("Graph Cache did not finish loading; Snapshot is not fully loaded yet.")
            return False
        j_filter["name"] = ["eq", "saveHistoricalData"]
        if (
            assurance_settings["disabled_historical_data"] is False
            and self._return_job_when_done(j_filter, retry=retry, timeout=timeout) is None
        ):
            logger.error("Historical Data did not finish loading; Snapshot is not fully loaded yet.")
            return False
        j_filter["name"] = ["eq", "report"]
        if (
            assurance_settings["disabled_intent_verification"] is False
            and self._return_job_when_done(j_filter, retry=retry, timeout=timeout) is None
        ):
            logger.error("Intent Calculations did not finish loading; Snapshot is not fully loaded yet.")
            return False
        return True



import os
from typing import Optional, List, Union, Dict
from uuid import UUID

from pydantic import BaseModel, Field

from .protocols import PROTOCOLS
from .trace import Trace

PYDANTIC_EXTRAS = os.getenv("IPFABRIC_PYDANTIC_EXTRAS", "allow")


class Extra(BaseModel, extra=PYDANTIC_EXTRAS):
    ip: Optional[str] = None


class Line(BaseModel, extra=PYDANTIC_EXTRAS):
    pattern: Optional[str] = None
    color: Optional[int] = None
    thickness: Optional[int] = None


class Style(BaseModel, extra=PYDANTIC_EXTRAS):
    background: Optional[str] = None
    line: Optional[Line] = None


class Position(BaseModel, extra=PYDANTIC_EXTRAS):
    x: Union[int, float]
    y: Union[int, float]


class EdgePosition(BaseModel, extra=PYDANTIC_EXTRAS):
    c: Position
    e: Position


class Label(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Optional[str] = None
    visible: Optional[bool] = None
    text: Optional[str] = None
    angle: Optional[int] = None
    anchor: Optional[Position] = None
    position: Optional[Position] = None


class Labels(BaseModel, extra=PYDANTIC_EXTRAS):
    center: Optional[Union[List[Label], Label]] = None
    source: Optional[Union[List[Label], Label]] = None
    target: Optional[Union[List[Label], Label]] = None


class ArrowHeads(BaseModel, extra=PYDANTIC_EXTRAS):
    target: Optional[List[Position]] = Field(default_factory=list)
    source: Optional[List[Position]] = Field(default_factory=list)


class Positions(BaseModel, extra=PYDANTIC_EXTRAS):
    line: Optional[List[Union[Position, EdgePosition]]] = Field(default_factory=list)
    arrowHeads: Optional[ArrowHeads] = None
    labels: Optional[Labels] = None


class Checks(BaseModel, extra=PYDANTIC_EXTRAS):
    green: int = Field(alias="0")
    blue: int = Field(alias="10")
    amber: int = Field(alias="20")
    red: int = Field(alias="30")


class Severity(Checks, extra=PYDANTIC_EXTRAS):
    pass


class Topics(BaseModel, extra=PYDANTIC_EXTRAS):
    acl: Checks = Field(alias="ACL")
    forwarding: Checks = Field(alias="FORWARDING")
    zonefw: Checks = Field(alias="ZONEFW")
    nat44: Optional[Checks] = Field(None, alias="NAT44")
    pbr: Optional[Checks] = Field(None, alias="PBR")


class TrafficScore(BaseModel, extra=PYDANTIC_EXTRAS):
    accepted: int
    dropped: int
    forwarded: int
    total: int


class Packets(BaseModel, extra=PYDANTIC_EXTRAS):
    packet: Optional[List[PROTOCOLS]] = Field(default_factory=list)
    ifaceName: Optional[str] = None
    prevEdgeIds: Optional[List[str]] = Field(default_factory=list)
    nextEdgeIds: Optional[List[str]] = Field(default_factory=list)
    severityInfo: Optional[Checks] = None
    trafficScore: Optional[TrafficScore] = None


class Node(BaseModel, extra=PYDANTIC_EXTRAS):
    path: Optional[str] = None
    boxId: Optional[str] = None
    children: List
    graphType: str
    id: str
    label: str
    parentPath: Optional[str] = None
    sn: str
    type: str
    stack: Optional[bool] = None
    position: Optional[Position] = None
    style: Optional[Style] = None
    acceptedPackets: Optional[Dict[str, Packets]] = Field(default_factory=dict)
    droppedPackets: Optional[Dict[str, Packets]] = Field(default_factory=dict)
    generatedPackets: Optional[Dict[str, Packets]] = Field(default_factory=dict)
    extra: Optional[Extra] = None


class Edge(BaseModel, extra=PYDANTIC_EXTRAS):
    direction: str
    source: str
    target: str
    edgeSettingsId: UUID
    id: str
    labels: Labels
    protocol: Optional[str] = ""
    shift: Optional[Union[int, float]] = None
    positions: Optional[Positions] = None
    style: Optional[Style] = None


class NetworkEdge(Edge, BaseModel, extra=PYDANTIC_EXTRAS):
    circle: bool
    children: List[str]


class PathLookupEdge(Edge, BaseModel, extra=PYDANTIC_EXTRAS):
    nextEdgeIds: List[str]
    prevEdgeIds: List[str]
    packet: List[PROTOCOLS]
    severityInfo: Severity
    sourceIfaceName: Optional[str] = None
    targetIfaceName: Optional[str] = None
    trafficScore: TrafficScore
    nextEdge: Optional[list] = Field(default_factory=list)
    prevEdge: Optional[list] = Field(default_factory=list)


class EventsSummary(BaseModel, extra=PYDANTIC_EXTRAS):
    flags: list
    topics: Topics
    global_list: list = Field(alias="global")


class Traces(BaseModel, extra=PYDANTIC_EXTRAS):
    severityInfo: Checks
    sourcePacketId: str
    targetPacketId: str
    trace: List[Trace]


class Decision(BaseModel, extra=PYDANTIC_EXTRAS):
    traces: List[Traces]
    trafficIn: Optional[Dict[str, List[str]]] = Field(default_factory=dict)
    trafficOut: Optional[Dict[str, List[str]]] = Field(default_factory=dict)


class Check(BaseModel, extra=PYDANTIC_EXTRAS):
    exists: bool


class PathLookup(BaseModel, extra=PYDANTIC_EXTRAS):
    eventsSummary: EventsSummary
    decisions: Dict[str, Decision]
    passingTraffic: str
    check: Check


class GraphResult(BaseModel, extra=PYDANTIC_EXTRAS):
    nodes: Dict[str, Node]
    edges: Dict[str, Union[NetworkEdge, PathLookupEdge]]
    pathlookup: Optional[PathLookup] = None



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ipfabric.client import EXPORT_FORMAT

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None

import re
import logging
from datetime import timedelta
from ipfabric.settings import Attributes
from httpx import HTTPStatusError
from typing import Optional, List, Union, Dict, DefaultDict, Set, overload, Literal, Any
from ipaddress import IPv4Interface
from collections import defaultdict
from case_insensitive_dict import CaseInsensitiveDict
from ipfabric.models.technology import Technology
from ipfabric.tools.shared import create_filter

from pydantic import BaseModel, Field, PrivateAttr
from ipfabric.exceptions import deprecated_args_decorator

logger = logging.getLogger("ipfabric")


class DeviceConfig(BaseModel):
    status: str
    current: Optional[str] = Field(None, alias="currentConfig")
    start: Optional[str] = Field(None, alias="startupConfig")


class Device(BaseModel):
    client: Optional[Any] = None
    attributes: Optional[dict] = None
    global_attributes: Optional[dict] = None
    domain: Optional[str] = None
    family: Optional[str] = None
    fqdn: Optional[str] = None
    hostname: str
    image: Optional[str] = None
    model: Optional[str] = None
    platform: Optional[str] = None
    processor: Optional[str] = None
    reload: Optional[str] = None
    sn: str
    uptime: Optional[timedelta] = None
    vendor: str
    version: Optional[str] = None
    blob_key: Optional[str] = Field(None, alias="blobKey")
    config_reg: Optional[str] = Field(None, alias="configReg")
    dev_type: str = Field(None, alias="devType")
    hostname_original: Optional[str] = Field(None, alias="hostnameOriginal")
    hostname_processed: Optional[str] = Field(None, alias="hostnameProcessed")
    login_ip: Optional[IPv4Interface] = Field(None, alias="loginIp")
    login_type: str = Field(None, alias="loginType")
    mem_total_bytes: Optional[float] = Field(None, alias="memoryTotalBytes")
    mem_used_bytes: Optional[float] = Field(None, alias="memoryUsedBytes")
    mem_utilization: Optional[float] = Field(None, alias="memoryUtilization")
    object_id: Optional[str] = Field(None, alias="objectId")
    routing_domain: Optional[int] = Field(None, alias="rd")
    site_name: str = Field(None, alias="siteName")
    sn_hw: str = Field(None, alias="snHw")
    stp_domain: Optional[int] = Field(None, alias="stpDomain")
    task_key: Optional[str] = Field(None, alias="taskKey")
    slug: Optional[str] = None

    def __repr__(self):
        return self.hostname

    def __str__(self):
        return self.hostname

    def __eq__(self, other):
        return self.sn == other.sn if isinstance(other, Device) else str(other)

    def __hash__(self):
        return hash(self.sn)

    @property
    def technology(self):
        return Technology(client=self.client, sn=self.sn)

    @property
    def site(self):
        return self.site_name

    @property
    def local_attributes(self):
        return self.attributes

    @classmethod
    def check_attribute(cls, attribute) -> True:
        if attribute not in cls.model_fields:
            raise AttributeError(f"Attribute {attribute} not in Device class.")
        return True

    @deprecated_args_decorator(version="6.8.0")
    def get_log_file(self, **kwargs) -> str:
        res = self.client.get("/os/logs/task/" + self.task_key)
        res.raise_for_status()
        return res.text

    @deprecated_args_decorator(version="6.8.0")
    def get_config(self, **kwargs) -> Union[None, DeviceConfig]:
        if not self.blob_key:
            logger.warning("Device Config not in Snapshot File. Please try using ipfabric.tools.DeviceConfigs")
            return None
        res = self.client.get("blobs/device-configuration/" + str(self.blob_key))
        res.raise_for_status()
        return DeviceConfig(**res.json())

    @deprecated_args_decorator(version="6.8.0")
    def interfaces(self, **kwargs) -> list:
        return self.client.inventory.interfaces.all(filters={"sn": ["eq", self.sn]})

    @deprecated_args_decorator(version="6.8.0")
    def pn(self, **kwargs) -> list:
        return self.client.inventory.pn.all(filters={"deviceSn": ["eq", self.sn]})

    @deprecated_args_decorator(version="6.8.0")
    def switchport(self, **kwargs) -> list:
        return self.technology.interfaces.switchport.all()

    @deprecated_args_decorator(version="6.8.0")
    def managed_ip_ipv4(self, **kwargs) -> list:
        return self.technology.addressing.managed_ip_ipv4.all()

    @deprecated_args_decorator(version="6.8.0")
    def managed_ip_ipv6(self, **kwargs) -> list:
        return self.technology.addressing.managed_ip_ipv6.all()

    @deprecated_args_decorator(version="6.8.0")
    def mac_table(self, **kwargs) -> list:
        return self.technology.addressing.mac_table.all()

    @deprecated_args_decorator(version="6.8.0")
    def arp_table(self, **kwargs) -> list:
        return self.technology.addressing.arp_table.all()

    @deprecated_args_decorator(version="6.8.0")
    def routes_ipv4(self, **kwargs) -> list:
        return self.technology.routing.routes_ipv4.all()

    @deprecated_args_decorator(version="6.8.0")
    def routes_ipv6(self, **kwargs) -> list:
        return self.technology.routing.routes_ipv6.all()

    @deprecated_args_decorator(version="6.8.0")
    def neighbors_all(self, **kwargs) -> list:
        return self.technology.neighbors.neighbors_all.all()

    def trigger_backup(self):
        return self.client.trigger_backup(sn=self.sn)

    @overload
    def fetch_all(
        self,
        client: Any,
        url: str,
        export: Literal["csv"],
        columns: List[str] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
    ) -> List[dict]: ...

    @overload
    def fetch_all(
        self,
        client: Any,
        url: str,
        export: Literal["csv"],
        columns: List[str] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch_all(
        self,
        client: Any,
        url: str,
        export: Literal["df"],
        columns: List[str] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
    ) -> DataFrame: ...

    @deprecated_args_decorator(version="6.8.0", arg_type="IPFClient", no_args=False)
    def fetch_all(
        self,
        client: Any,
        url: str,
        export: EXPORT_FORMAT = "json",
        columns: List[str] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from IP Fabric for specified endpoint filtered on the `sn` of the device

        Args:
            url: Example tables/vlan/device-summary
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional list of columns to return, None will return all
            filters: Optional dictionary of filters which will be merged with the sn filter
            snapshot_id: Optional snapshot_id to override default
            reports: String of frontend URL where the reports are displayed or a list of report IDs
            sort: Optional dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], bytes, pandas.DataFrame]: List of dict if json, bytes string if CSV, DataFrame is df
        """
        all_columns, f = create_filter(self.client, url, filters, self.sn)
        return self.client.fetch_all(
            url,
            filters=f,
            columns=columns or all_columns,
            export=export,
            snapshot_id=snapshot_id,
            reports=reports,
            sort=sort,
            csv_tz=csv_tz,
        )


class DeviceDict(CaseInsensitiveDict):
    """CaseInsensitiveDict with functions to search or regex on dictionary keys."""

    def __init__(self, attribute, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attribute = attribute

    @staticmethod
    def _new_dict(a):
        return DeviceDict[str, Device](attribute=a) if a == "sn" else DeviceDict[str, List[Device]](attribute=a)

    @overload
    def regex(self: DeviceDict[str, List[Device]], pattern: str, *flags: int) -> DeviceDict[str, List[Device]]: ...

    @overload
    def regex(self: DeviceDict[str, Device], pattern: str, *flags: int) -> DeviceDict[str, Device]: ...

    def regex(self, pattern: str, *flags: int) -> DeviceDict[str, Union[List[Device], Device]]:
        """
        Case-sensitive regex search on dictionary keys.
        Args:
            pattern: str: Regex string to search.
            *flags: int or re.RegexFlag: Regex flags to use.

        Returns:
            DeviceDict: New instance of DeviceDict (CaseInsensitiveDict)
        """
        regex = re.compile(pattern, flags=sum(flags))
        new_dict = self._new_dict(self.attribute)
        [new_dict.update({key: value}) for key, value in self._data.values() if key and regex.search(key)]
        return new_dict

    @overload
    def search(
        self: DeviceDict[str, List[Device]], pattern: Union[List[str], str]
    ) -> DeviceDict[str, List[Device]]: ...

    @overload
    def search(self: DeviceDict[str, Device], pattern: Union[List[str], str]) -> DeviceDict[str, Device]: ...

    def search(self, pattern: Union[List[str], str]) -> DeviceDict[str, Union[List[Device], Device]]:
        """
        Case-insensitive search on dictionary keys.
        Args:
            pattern: Union[List[str], str]: String or List of strings to match on.

        Returns:
            DeviceDict: New instance of DeviceDict (CaseInsensitiveDict)
        """
        pattern = [pattern.lower()] if isinstance(pattern, str) else [p.lower() for p in pattern]
        new_dict = self._new_dict(self.attribute)
        [new_dict.update({o_key: value}) for key, (o_key, value) in self._data.items() if key in pattern]
        return new_dict

    def _flatten_devs(self) -> List[Device]:
        devices = list()
        for value in self.values():
            if isinstance(value, list):
                devices.extend(value)
            else:
                devices.append(value)
        return devices

    def sub_search(self, attribute: str, pattern: Union[List[str], str]) -> DeviceDict[str, List[Device]]:
        """
        Case-insensitive sub search of another Device attribute.
        Args:
            attribute: str: Attribute of Device class.
            pattern: Union[List[str], str]: String or List of strings to match on.

        Returns:
             DeviceDict: New instance of DeviceDict (CaseInsensitiveDict) grouped by new attribute.
        """
        Device.check_attribute(attribute)
        return Devices.group_dev_by_attr(self._flatten_devs(), attribute).search(pattern)

    def sub_regex(self, attribute: str, pattern: Union[List[str], str], *flags: int) -> DeviceDict[str, List[Device]]:
        """
        Case-sensitive regex sub search of another Device attribute.
        Args:
            attribute: str: Attribute of Device class.
            pattern: str: Regex string to search.
            *flags: int or re.RegexFlag: Regex flags to use.

        Returns:
            DeviceDict: New instance of DeviceDict (CaseInsensitiveDict) grouped by new attribute.
        """
        Device.check_attribute(attribute)
        return Devices.group_dev_by_attr(self._flatten_devs(), attribute).regex(pattern, *flags)


class Devices(BaseModel):
    snapshot_id: str
    client: Any
    _attrs: Optional[DefaultDict[str, Set[str]]] = PrivateAttr()
    _global_attrs: Optional[DefaultDict[str, Set[str]]] = PrivateAttr()
    _all: List[Device] = PrivateAttr()

    def __init__(
        self,
        snapshot_id: str,
        client: Any,
        devices: Optional[List[dict]] = None,
        device_filters: Optional[dict] = None,
        device_attr_filters: Optional[dict] = None,
    ):
        super().__init__(snapshot_id=snapshot_id, client=client)
        self.update(devices, device_filters, device_attr_filters)

    def update(self, devices: List[dict] = None, device_filters: dict = None, device_attr_filters: dict = None):
        devices = devices if devices else self._load_devices(device_filters, device_attr_filters)
        if not devices:
            return None
        try:
            self._attrs, lcl_attr = self._parse_attrs(
                Attributes(client=self.client, snapshot_id=self.client.snapshot_id).all()
            )
        except HTTPStatusError:
            logger.warning(
                self.client._api_insuf_rights
                + 'on POST "/tables/snapshot-attributes". Cannot load Local (snapshot) Attributes in Devices.'
            )
            lcl_attr = dict()
        try:
            self._global_attrs, glb_attr = self._parse_attrs(Attributes(client=self.client).all())
        except HTTPStatusError:
            logger.warning(
                self.client._api_insuf_rights
                + 'on POST "/tables/global-attributes". Cannot load Global Attributes in Devices.'
            )
            glb_attr = dict()
        try:
            blob_keys = {
                b["sn"]: b["blobKey"]
                for b in self.client.fetch_all("/tables/management/configuration/saved", columns=["sn", "blobKey"])
            }
        except HTTPStatusError:
            logger.warning(
                self.client._api_insuf_rights + 'on POST "/tables/management/configuration/saved". '
                "You will not be able to pull device config from Device model."
            )
            blob_keys = dict()
        self._all = [
            Device(
                **d,
                attributes=lcl_attr.get(d["sn"], dict()),
                global_attributes=glb_attr.get(d["sn"], dict()),
                blobKey=blob_keys.get(d["sn"], None),
                client=self.client,
            )
            for d in devices
        ]

    def _load_devices(self, device_filters: dict = None, device_attr_filters: dict = None):
        if self.client._no_loaded_snapshots:
            logger.warning("No loaded snapshots, cannot load devices.")
        else:
            if not device_attr_filters and self.client.attribute_filters:
                logger.warning(
                    f"Global `attribute_filters` is set; only pulling devices matching:\n{self.client.attribute_filters}."
                )
            try:
                return self.client.inventory.devices.all(filters=device_filters, attr_filters=device_attr_filters)
            except HTTPStatusError:
                logger.warning(self._api_insuf_rights + 'on POST "/tables/inventory/devices". Will not load Devices.')
        return list()

    @staticmethod
    def _parse_attrs(attributes: List[dict] = None):
        if not attributes:
            return None, dict()
        cls_attr, dev_attr = defaultdict(set), defaultdict(dict)
        for d in attributes:
            dev_attr[d["sn"]].update({d["name"]: d["value"]})
            cls_attr[d["name"]].add(d["value"])
        return cls_attr, dev_attr

    @property
    def all(self) -> List[Device]:
        """Returns List[Device]."""
        return self._all

    def _check_attr_name(self, name: str) -> bool:
        if self._attrs is None:
            logger.warning("Attributes were not loaded into devices.")
        elif name not in self._attrs:
            logger.warning(f'Attribute key "{name}" not found in snapshot "{self.snapshot_id}".')
        else:
            return True
        return False

    def _filter_attr(self, devs: Set[Device], name: str, values: Union[List[str], str]) -> Set[Device]:
        if not self._check_attr_name(name):
            return set()
        values = values if isinstance(values, list) else [values]
        for dev in devs.copy():
            dev_attr = dev.attributes.get(name, None)
            if not dev_attr or dev_attr not in values:
                devs.discard(dev)
        return devs

    def filter_by_attr(self, name: str, values: Union[List[str], str]) -> List[Device]:
        """
        Return list of devices with an attribute set to a value
        Args:
            name: str: Attribute name
            values: Union[List[str], str]: Single attribute value or list of values to match.
        Returns:
            List[Device]
        """
        return list(self._filter_attr(set(self.all.copy()), name, values))

    def filter_by_attrs(self, attr_filter: Dict[str, Union[List[str], str]]) -> List[Device]:
        """
        Return list of devices matching multiple key/value attribute pairs
        Args:
            attr_filter: dict: {'ATTR_1': 'VALUE_1', 'ATTR_2': ['VALUE_2', 'VALUE_3']}
        Returns:
            List[Device]
        """
        devs = set(self.all.copy())
        for k, v in attr_filter.items():
            devs = self._filter_attr(devs, k, v)
        return list(devs)

    def has_attr(self, name: str) -> List[Device]:
        """
        Return list of devices that has an attribute set matching name.
        Args:
            name: str: Attribute name
        Returns:
            List[Device]
        """
        return [d for d in self.all if d.attributes.get(name, None)] if self._check_attr_name(name) else list()

    def does_not_have_attr(self, name: str) -> List[Device]:
        """
        Return list of devices that does not have an attribute set matching name.
        Args:
            name: str: Attribute name
        Returns:
            List[Device]
        """
        return [d for d in self.all if not d.attributes.get(name, None)] if self._check_attr_name(name) else list()

    def _group_dev_by_attr(self, attribute: str) -> DeviceDict[str, List[Device]]:
        return self.group_dev_by_attr(self._all, attribute)

    @classmethod
    def group_dev_by_attr(cls, devices: List[Device], attribute: str) -> DeviceDict[str, List[Device]]:
        devs = defaultdict(list)
        [devs[getattr(d, attribute)].append(d) for d in devices]
        return DeviceDict(attribute=attribute, data=devs)

    @property
    def by_sn(self) -> DeviceDict[str, Device]:
        """Returns Case-insensitive DeviceDict {'sn': Device}."""
        return DeviceDict(attribute="sn", data={d.sn: d for d in self._all})

    @property
    def by_hostname_original(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'hostname': [Device]}."""
        return self._group_dev_by_attr("hostname_original")

    @property
    def by_hostname(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'hostname': [Device]}."""
        return self._group_dev_by_attr("hostname")

    @property
    def by_sn_hw(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'sn_hw': [Device]}."""
        return self._group_dev_by_attr("sn_hw")

    @property
    def by_site(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'site': [Device]}."""
        return self._group_dev_by_attr("site")

    @property
    def by_vendor(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'vendor': [Device]}."""
        return self._group_dev_by_attr("vendor")

    @property
    def by_family(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'family': [Device]}."""
        return self._group_dev_by_attr("family")

    @property
    def by_platform(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'platform': [Device]}."""
        return self._group_dev_by_attr("platform")

    @property
    def by_model(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'model': [Device]}."""
        return self._group_dev_by_attr("model")

    @property
    def by_version(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'version': [Device]}."""
        return self._group_dev_by_attr("version")

    @property
    def by_fqdn(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'version': [Device]}."""
        return self._group_dev_by_attr("fqdn")

    @property
    def by_login_type(self) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'version': [Device]}."""
        return self._group_dev_by_attr("login_type")

    def by_custom(self, attribute) -> DeviceDict[str, List[Device]]:
        """Returns Case-insensitive DeviceDict {'version': [Device]}."""
        Device.check_attribute(attribute)
        return self._group_dev_by_attr(attribute)



import logging
from typing import Any

from pydantic import BaseModel

from .table import Table, DeviceTable

logger = logging.getLogger("ipfabric")

IGNORE_COLUMNS = {"id"}


class Inventory(BaseModel):
    """model for inventories"""

    client: Any = None

    @property
    def sites(self):
        return Table(client=self.client, endpoint="tables/inventory/sites")

    @property
    def vendors(self):
        return Table(client=self.client, endpoint="tables/inventory/summary/vendors")

    @property
    def devices(self):
        return DeviceTable(client=self.client)

    @property
    def models(self):
        return Table(client=self.client, endpoint="tables/inventory/summary/models")

    @property
    def os_version_consistency(self):
        return Table(client=self.client, endpoint="tables/management/osver-consistency")

    @property
    def eol_summary(self):
        return Table(client=self.client, endpoint="tables/reports/eof/summary")

    @property
    def eol_details(self):
        return Table(client=self.client, endpoint="tables/reports/eof/detail")

    @property
    def platforms(self):
        return Table(client=self.client, endpoint="tables/inventory/summary/platforms")

    @property
    def pn(self):
        return Table(client=self.client, endpoint="tables/inventory/pn")

    @property
    def families(self):
        return Table(client=self.client, endpoint="tables/inventory/summary/families")

    @property
    def interfaces(self):
        return Table(client=self.client, endpoint="tables/inventory/interfaces")

    @property
    def hosts(self):
        return Table(client=self.client, endpoint="tables/addressing/hosts")

    @property
    def hosts_ipv6(self):
        return Table(client=self.client, endpoint="tables/addressing/ipv6-hosts")

    @property
    def phones(self):
        return Table(client=self.client, endpoint="tables/inventory/phones")

    @property
    def fans(self):
        return Table(client=self.client, endpoint="tables/inventory/fans")

    @property
    def modules(self):
        return Table(client=self.client, endpoint="tables/inventory/modules")



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Cloud(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def virtual_machines(self):
        return models.Table(client=self.client, endpoint="tables/cloud/virtual-machines", sn=self.sn)

    @property
    def virtual_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/cloud/virtual-machines-interfaces", sn=self.sn)



import logging
from datetime import datetime
from ipaddress import IPv4Network
from typing import Optional, Any, Union, ClassVar

from dateutil import parser
from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass

logger = logging.getLogger("ipfabric")


class Expiration(BaseModel):
    enabled: bool
    value: Optional[datetime] = None


class Credential(BaseModel):
    username: str
    network: list
    excluded: list = Field(None, alias="excludeNetworks")
    expiration: Expiration = Field(None, alias="expirationDate")
    credential_id: str = Field(None, alias="id")
    encrypt_password: str = Field(None, alias="password")
    priority: Optional[int] = None
    config_mgmt: bool = Field(None, alias="syslog")
    notes: Optional[str] = None


class Privilege(BaseModel):
    network: list = Field(None, alias="includeNetworks")
    excluded: list = Field(None, alias="excludeNetworks")
    expiration: Expiration = Field(None, alias="expirationDate")
    privilege_id: str = Field(None, alias="id")
    encrypt_password: str = Field(None, alias="password")
    priority: Optional[int] = None
    username: str
    notes: Optional[str] = None


@dataclass
class Authentication:
    client: Any
    credentials: dict = Field(default=dict())
    enables: dict = Field(default=dict())
    _cred_url: ClassVar[str] = "settings/credentials"
    _priv_url: ClassVar[str] = "settings/privileges"

    def __post_init__(self):
        self.get_credentials()
        self.get_enables()

    def get_credentials(self) -> dict:
        """
        Get all credentials and sets them in the Authentication.credentials
        :return: self.credentials
        """
        res = self.client.get(self._cred_url)
        res.raise_for_status()
        self.credentials = {cred["priority"]: Credential(**cred) for cred in res.json()["data"]}
        return self.credentials

    def get_enables(self) -> dict:
        """
        Get all privileges (enable passwords) and sets them in the Authentication.enables
        :return:
        """
        res = self.client.get(self._priv_url)
        res.raise_for_status()
        self.enables = {priv["priority"]: Privilege(**priv) for priv in res.json()["data"]}
        return self.enables

    def _create_payload(self, username, password, notes, network, excluded, expiration):
        networks = network or ["0.0.0.0/0"]
        excluded = excluded or list()
        if expiration:
            expires = dict(
                enabled=True,
                value=parser.parse(expiration).strftime("%Y-%m-%d %H:%M:%S"),
            )
        else:
            expires = dict(enabled=False)
        payload = {
            "password": password,
            "username": username,
            "notes": notes or username,
            "excludeNetworks": self._check_networks(excluded),
            "expirationDate": expires,
            "network": self._check_networks(networks),
        }
        return payload

    def create_credential(
        self,
        username: str,
        password: str,
        networks: list = None,
        notes: str = None,
        excluded: list = None,
        config_mgmt: bool = False,
        expiration: str = None,
    ) -> Credential:
        """
        Creates a new credential. Requires username and password and will default to all networks with no expiration.
        Does not default to use for configuration management, please set to true if needed.
        After creation Authentication.credentials will be updated with new priorities and the new cred is returned.
        :param username: str: Username
        :param password: str: Unencrypted password
        :param networks: list: List of networks defaults to ["0.0.0.0/0"]
        :param notes: str: Optional Note/Description of credential
        :param excluded: list: Optional list of networks to exclude
        :param config_mgmt: bool: Default False - do not use for configuration management
        :param expiration: str: Optional date for expiration, if none then do not expire.
                                To ensure correct date use YYYYMMDD or MM/DD/YYYY formats
        :return: Credential: Obj: Credential Obj with ID and encrypted password
        """
        payload = self._create_payload(username, password, notes, networks, excluded, expiration)
        payload.update({"syslog": config_mgmt})
        res = self.client.post(self._cred_url, json=payload)
        res.raise_for_status()
        self.get_credentials()
        cred = Credential(**res.json())
        logger.info(f"Created credential with username {cred.username} and ID of {cred.credential_id}")
        return cred

    def create_enable(
        self,
        username: str,
        password: str,
        networks: list = None,
        notes: str = None,
        excluded: list = None,
        expiration: str = None,
    ) -> Privilege:
        """
        Creates a new enable password (privilege account).
        Requires username and password and will default to all networks with no expiration.
        After creation Authentication.enables will be updated with new priorities and the new enable is returned.
        :param username: str: Username
        :param password: str: Unencrypted password
        :param networks: list: List of networks defaults to ["0.0.0.0/0"]
        :param notes: str: Optional Note/Description of credential
        :param excluded: list: Optional list of networks to exclude
        :param expiration: str: Optional date for expiration, if none then do not expire.
                                To ensure correct date use YYYYMMDD or MM/DD/YYYY formats
        :return: Privilege: Obj: Privilege Obj with ID and encrypted password
        """
        payload = self._create_payload(username, password, notes, networks, excluded, expiration)
        payload["includeNetworks"] = payload.pop("network")
        res = self.client.post(self._priv_url, json=payload)
        res.raise_for_status()
        self.get_enables()
        priv = Privilege(**res.json())
        logger.info(f"Created enable password with username {priv.username} and ID of {priv.privilege_id}")
        return priv

    def delete_credential(self, credential: Union[Credential, str]) -> None:
        """
        Deletes a credential and updates Authentication.credentials with new priorities.
        :param credential: Union[Credential, str]: Cred ID in a string or Credential object
        :return:
        """
        cred = credential.credential_id if isinstance(credential, Credential) else credential
        res = self.client.request("DELETE", self._cred_url, json=[cred])
        res.raise_for_status()
        self.get_credentials()
        logger.warning(f"Deleted credential ID {cred}")

    def delete_enable(self, enable: Union[Privilege, str]) -> None:
        """
        Deletes an enable password (privilege account) and updates Authentication.enable with new priorities.
        :param enable: Union[Privilege, str]: Enable ID in a string or Privilege object
        :return:
        """
        priv = enable.privilege_id if isinstance(enable, Privilege) else enable
        res = self.client.request("DELETE", self._priv_url, json=[priv])
        res.raise_for_status()
        self.get_enables()
        logger.warning(f"Deleted enable password ID {priv}")

    def update_cred_priority(self, credentials: dict) -> dict:
        """
        Updates the priority of credentials.  Reorder Authentication.credentials dictionary and submit to this method.
        :param credentials: dict: {priority: Credential}
        :return: self.credentials: dict: {priority: Credential}
        """
        payload = [dict(id=c.credential_id, priority=p) for p, c in credentials.items()]
        res = self.client.patch(self._cred_url, json=payload)
        res.raise_for_status()
        self.get_credentials()
        return self.credentials

    def update_enable_priority(self, enables: dict) -> dict:
        """
        Updates the priority of enable passwords.  Reorder Authentication.enables dictionary and submit to this method.
        :param enables: dict: {priority: Privilege}
        :return: self.enables: dict: {priority: Privilege}
        """
        payload = [dict(id=e.privilege_id, priority=p) for p, e in enables.items()]
        res = self.client.patch(self._priv_url, json=payload)
        res.raise_for_status()
        self.get_enables()
        return self.enables

    @staticmethod
    def _check_networks(subnets):
        return [IPv4Network(network).with_prefixlen for network in subnets]



import os
from typing import Optional, List, Union, Literal

try:
    from typing import Annotated  # py38 required Annotated
except ImportError:
    from typing_extensions import Annotated

from pydantic import BaseModel, Field

from .protocols import PROTOCOLS

PYDANTIC_EXTRAS = os.getenv("IPFABRIC_PYDANTIC_EXTRAS", "allow")


class SeverityInfo(BaseModel, extra=PYDANTIC_EXTRAS):
    name: str
    severity: int
    topic: str
    details: Optional[List[str]] = Field(default_factory=list)


class PacketDataMatch(BaseModel, extra=PYDANTIC_EXTRAS):
    field: str
    value: Union[str, List[str], None]
    type: Literal["packet data match"]


class RemoveHeader(BaseModel, extra=PYDANTIC_EXTRAS):
    index: int
    headerType: str
    type: Literal["remove header"]


class Filter(BaseModel, extra=PYDANTIC_EXTRAS):
    label: Optional[int] = None
    mask: Optional[int] = None
    vrf: Optional[str] = None
    prefix: Optional[str] = None
    ip: Optional[Union[str, List[str]]] = None
    groupIp: Optional[str] = None
    sourceIp: Optional[str] = None
    inIfaceName: Optional[str] = None
    mac: Optional[str] = None
    vlanNum: Optional[int] = None
    name: Optional[str] = None
    port: Optional[str] = None
    network: Optional[str] = None
    vrfVxlanId: Optional[int] = None


class TableEntry(BaseModel, extra=PYDANTIC_EXTRAS):
    filter: Filter
    table: str


class TableEntryMatch(TableEntry, BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["table entry match"]


class TableEntryNotFound(TableEntry, BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["table entry not found"]


class InsertHeader(BaseModel, extra=PYDANTIC_EXTRAS):
    header: PROTOCOLS
    headerType: str
    index: int
    type: Literal["insert header"]


class Patch(BaseModel, extra=PYDANTIC_EXTRAS):
    stack: Optional[List[int]] = None
    ttl: Optional[int] = None
    dst: Optional[List[str]] = None
    policyApplied: Optional[bool] = None


class PatchHeader(BaseModel, extra=PYDANTIC_EXTRAS):
    patch: Patch
    index: int
    type: Literal["patch header"]
    headerType: Optional[str] = None


class DropPacket(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["drop packet"]
    reason: str
    severityInfo: Optional[SeverityInfo] = None
    ips: Optional[List[str]] = None
    groupIp: Optional[List[str]] = None
    sourceIp: Optional[List[str]] = None


class SeverityEvent(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["severity event"]
    severityInfo: Optional[SeverityInfo] = None


class BaseEvent(BaseModel, extra=PYDANTIC_EXTRAS):
    decidingPolicyName: Optional[str] = None
    decidingRule: Optional[List[int]] = None
    protocolId: str
    securityType: str
    severityInfo: Optional[SeverityInfo] = None


class SecurityCheck(BaseEvent, extra=PYDANTIC_EXTRAS):
    type: Literal["security check"]
    isInvisible: Optional[bool] = None
    isDefaultRule: Optional[bool] = None
    allHidden: Optional[bool] = None


class SecurityCheckIgnored(SecurityCheck, BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["security check ignored"]


class VirtualRouting(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["virtual routing"]
    ifaceName: str


class AcceptPacket(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["accept packet"]


class DestinationNAT(BaseEvent, BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["destination NAT"]
    field: Optional[str] = None


class SourceNAT(BaseEvent, BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["source NAT"]
    field: Optional[str] = None


class PolicyBasedRouting(BaseEvent, BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["policy based routing"]
    field: Optional[str] = None


class FloodPacket(SeverityEvent, BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["flood packet"]
    reason: str


EVENT = Annotated[
    Union[
        PacketDataMatch,
        RemoveHeader,
        TableEntryMatch,
        TableEntryNotFound,
        VirtualRouting,
        AcceptPacket,
        InsertHeader,
        PatchHeader,
        DropPacket,
        SeverityEvent,
        SecurityCheckIgnored,
        SecurityCheck,
        DestinationNAT,
        PolicyBasedRouting,
        SourceNAT,
        FloodPacket,
    ],
    Field(discriminator="type"),
]


class Trace(BaseModel, extra=PYDANTIC_EXTRAS):
    chain: str
    phase: str
    events: List[EVENT]



import logging
import re
from typing import Optional, Union, Generator

import httpx
from pydantic import field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from ipfabric.tools import VALID_REFS

logger = logging.getLogger("ipfabric")


class AccessToken(httpx.Auth):
    def __init__(self, client: httpx.Client):
        self.client = client

    def auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        response = yield request

        if response.status_code == 401:
            response.read()
            if "API_EXPIRED_ACCESS_TOKEN" in response.text:
                resp = self.client.post("/api/auth/token")  # Use refreshToken in Cookies to get new accessToken
                resp.raise_for_status()  # Response updates accessToken in shared CookieJar
                request.headers["Cookie"] = "accessToken=" + self.client.cookies["accessToken"]  # Update request
                yield request
        return response


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", env_prefix="ipf_", extra="allow")
    url: Optional[str] = None
    version: Optional[Union[int, float, str]] = None
    token: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    snapshot: Optional[str] = None
    verify: Union[bool, int, str] = True
    timeout: Optional[float] = None

    @field_validator("version")
    @classmethod
    def _valid_version(cls, v: Union[None, int, float, str]) -> Union[None, str]:
        if v and isinstance(v, (int, float)):
            v = "v" + str(v)
        if not v or re.match(r"v\d(\.\d)?", v):
            return v
        else:
            raise ValueError(f"IPF_VERSION ({v}) is not valid, must be like `v#` or `v#.#`.")

    @field_validator("snapshot")
    @classmethod
    def _valid_snapshot(cls, v: Union[None, str]) -> Union[None, str]:
        if v is None or v in VALID_REFS:
            return v
        elif re.match(r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$", v.lower()):
            return v.lower()
        else:
            raise ValueError(f"IPF_SNAPSHOT ({v}) is not valid, must be a UUID or one of {VALID_REFS}.")

    @field_validator("verify")
    @classmethod
    def _verify(cls, v: Union[bool, int, str]) -> Union[bool, str]:
        if isinstance(v, bool):
            return v
        if v.lower() in {0, "0", "off", "f", "false", "n", "no", 1, "1", "on", "t", "true", "y", "yes"}:
            return False if v.lower() in {0, "0", "off", "f", "false", "n", "no"} else True
        else:
            return v

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Needed for context"""
        pass



import logging
from typing import Any, Optional

import pytz
from pydantic import Field, BaseModel

logger = logging.getLogger("ipfabric")


class Token(BaseModel):
    token_id: str = Field(None, alias="id")
    description: Optional[str] = None
    role_ids: list = Field(None, alias="roleIds")


class User(BaseModel):
    username: str
    user_id: str = Field(None, alias="id")
    local: Optional[bool] = Field(None, alias="isLocal")
    sso_provider: Optional[Any] = Field(None, alias="ssoProvider")
    domains: Optional[Any] = Field(None, alias="domainSuffixes")
    role_names: Optional[list] = Field(alias="roleNames", default_factory=list)
    role_ids: list = Field(None, alias="roleIds")
    ldap_id: Optional[Any] = Field(None, alias="ldapId")
    timezone: Optional[str] = None
    token: Optional[Token] = None
    _get_snapshots_settings: Optional[bool] = None

    @property
    def is_admin(self):
        return (
            True
            if (self.token and "admin" in self.token.role_ids) or (not self.token and "admin" in self.role_ids)
            else False
        )

    @property
    def get_snapshots_settings(self):
        """If the User/Token has access to snapshots/:key/settings Endpoint."""
        if self._get_snapshots_settings is None and self.is_admin:
            self._get_snapshots_settings = True
        return self._get_snapshots_settings

    def set_snapshots_settings(self, v: bool):
        self._get_snapshots_settings = v

    @property
    def error_msg(self):
        msg = f'User "{self.username}" '
        if self.token:
            msg += f'Token "{self.token.description}" '
        return msg


class Role(BaseModel):
    name: str
    role_id: str = Field(None, alias="id")
    description: Optional[str] = None
    role_type: str = Field(None, alias="type")
    admin: bool = Field(None, alias="isAdmin")
    system: bool = Field(None, alias="isSystem")


class UserMgmt:
    def __init__(self, client):
        self.client: Any = client
        self.roles = self.get_roles()
        self.users = self.get_users()

    def get_roles(self, role_name: str = None):
        """
        Gets all users or filters on one of the options.
        :param role_name: str: Role Name to filter
        :return: List of roles
        """
        filters = {"name": ["ieq", role_name]} if role_name else None
        return [Role(**role) for role in self.client.fetch_all("tables/roles", filters=filters, snapshot=False)]

    @property
    def roles_by_id(self):
        return {r.role_id: r for r in self.roles}

    @property
    def roles_by_name(self):
        return {r.name: r for r in self.roles}

    def get_users(self, username: str = None):
        """
        Gets all users or filters on one of the options.
        :param username: str: Username to filter
        :return: List of users
        """
        filters = {"username": ["ieq", username]} if username else None
        users = self.client.fetch_all("tables/users", filters=filters, snapshot=False)
        return [User(**user) for user in users]

    def get_user_by_id(self, user_id: str):
        """
        Gets a user by ID
        :param user_id: Union[str, int]: User ID to filter
        :return: User
        """
        resp = self.client.get("users/" + str(user_id))
        resp.raise_for_status()
        user = resp.json()
        return User(roleNames=[self.roles_by_id[r].name for r in user["roleIds"]], **user)

    def add_user(
        self,
        username: str,
        password: str,
        roles: list,
        timezone: str = "UTC",
    ):
        """
        Adds a user
        :param username: str: Username
        :param password: str: Must be 8 characters
        :param roles: list: Role IDs for Users
        :param timezone: str: v4.2 and above, Defaults UTC.  See pytz.all_timezones for correct syntax
        :return: User
        """
        if len(password) < 8:
            raise SyntaxError("Password must be 8 characters.")
        if not all(x in [r.role_id for r in self.roles] for x in roles):
            raise SyntaxError(f"Only accepted scopes are {[r.role_id for r in self.roles]}")
        payload = {
            "username": username,
            "password": password,
            "roleIds": roles,
        }
        if timezone not in pytz.all_timezones:
            raise ValueError(
                f"Timezone {timezone} is not located. This is case sensitive please see pytz.all_timezones."
            )
        payload["timezone"] = timezone
        resp = self.client.post("users", json=payload)
        resp.raise_for_status()
        user = resp.json()
        return User(roleNames=[self.roles_by_id[r].name for r in user["roleIds"]], **user)

    def delete_user(self, user_id: str):
        """
        Deletes a user and returns list of remaining users
        :param user_id:
        :return:
        """
        resp = self.client.delete("users/" + str(user_id))
        resp.raise_for_status()
        return self.get_users()



from .api_tokens import APIToken
from .attributes import Attributes
from .authentication import Authentication
from .seeds import Seeds
from .site_separation import SiteSeparation
from .user_mgmt import UserMgmt
from .vendor_api import VendorAPI
from .vendor_api_models import (
    AWS_REGIONS,
    AWS,
    Azure,
    CheckPointApiKey,
    CheckPointUserAuth,
    CiscoAPIC,
    CiscoFMC,
    ForcePoint,
    GCP,
    JuniperMist,
    Merakiv1,
    Prisma,
    RuckusVirtualSmartZone,
    SilverPeak,
    Versa,
    Viptela,
    NSXT,
)

__all__ = [
    "APIToken",
    "Attributes",
    "Authentication",
    "Seeds",
    "SiteSeparation",
    "UserMgmt",
    "VendorAPI",
    "AWS_REGIONS",
    "AWS",
    "Azure",
    "CheckPointApiKey",
    "CheckPointUserAuth",
    "CiscoAPIC",
    "CiscoFMC",
    "ForcePoint",
    "GCP",
    "JuniperMist",
    "Merakiv1",
    "Prisma",
    "RuckusVirtualSmartZone",
    "SilverPeak",
    "Versa",
    "Viptela",
    "NSXT",
]



import logging
from typing import Any, Union, Optional, List

from pydantic import Field, BaseModel

from ipfabric.settings.user_mgmt import UserMgmt
from ipfabric.tools.shared import date_parser

logger = logging.getLogger("ipfabric")


class Token(BaseModel):
    description: str
    usage: int
    username: str
    expires: Optional[int] = None
    token: Optional[str] = None
    user_id: str = Field(alias="userId")
    user_role_ids: Optional[List[str]] = Field(None, alias="userRoleIds")
    token_id: str = Field(alias="id")
    is_expired: bool = Field(alias="isExpired")
    last_used: Optional[Union[str, int]] = Field(None, alias="lastUsed")
    role_ids: List[str] = Field(alias="roleIds", default_factory=list)
    role_names: Optional[List[str]] = Field(None, alias="roleNames")


class APIToken:
    def __init__(self, client):
        self.client: Any = client
        self.tokens = self.get_tokens()

    @property
    def tokens_by_id(self):
        return {t.token_id: t for t in self.tokens}

    @property
    def tokens_by_description(self):
        return {t.description: t for t in self.tokens}

    def get_tokens(self):
        res = self.client.get("api-tokens")
        res.raise_for_status()
        return [Token(**t) for t in res.json()]

    def _check_roles(self, role_ids: list, role_names: list):
        if not role_names and not role_ids:
            raise SyntaxError("No Role Ids or Names provided.")
        roles = UserMgmt(self.client)
        checked_roles = list()
        for role in role_ids:
            if str(role) in roles.roles_by_id:
                checked_roles.append(str(role))
            else:
                raise ValueError(f"Role ID '{role}' does not exist.")
        for role in role_names:
            if str(role) in roles.roles_by_name:
                checked_roles.append(str(role))
            else:
                raise ValueError(f"Role Name '{role}' does not exist.")
        return checked_roles

    def add_token(
        self,
        descr: str,
        role_ids: Optional[list] = None,
        role_names: Optional[list] = None,
        expires: Optional[Union[str, int]] = None,
    ):
        payload = dict(
            description=descr,
            expires=int(date_parser(expires).timestamp() * 1000) if expires else None,
            roleIds=self._check_roles(role_ids or list(), role_names or list()),
        )
        res = self.client.post("api-tokens", json=payload)
        res.raise_for_status()
        return Token(**res.json())

    def delete_token(self, token_id: [int, str, Token]):
        token_id = token_id.token_id if isinstance(token_id, Token) else str(token_id)
        if token_id not in self.tokens_by_id:
            raise ValueError(f"Could not find token matching ID '{token_id}'.")
        res = self.client.delete("api-tokens/" + token_id)
        res.raise_for_status()
        self.tokens = self.get_tokens()
        return self.tokens

    def delete_token_by_description(self, token_description: str):
        if token_description not in self.tokens_by_description:
            raise ValueError(f"Could not find token matching description '{token_description}'.")
        return self.delete_token(self.tokens_by_description[token_description].token_id)



from .constants import VALID_NET_PROTOCOLS, VALID_PATH_PROTOCOLS
from .graph_parameters import (
    Unicast,
    Multicast,
    Host2GW,
    Network,
    OtherOptions,
    Algorithm,
    EntryPoint,
    Layout,
)
from .graph_settings import NetworkSettings, PathLookupSettings, Overlay

__all__ = [
    "Unicast",
    "Multicast",
    "Host2GW",
    "Network",
    "OtherOptions",
    "Algorithm",
    "Overlay",
    "NetworkSettings",
    "PathLookupSettings",
    "EntryPoint",
    "VALID_NET_PROTOCOLS",
    "VALID_PATH_PROTOCOLS",
    "Layout",
]



from typing import Optional, List

from pydantic import BaseModel


class OAS(BaseModel):
    api_endpoint: str
    web_endpoint: Optional[str] = None
    columns: Optional[List[str]] = None
    nested_columns: Optional[List[str]] = None



import os
from typing import List, Union, Optional, Literal

from pydantic import BaseModel, Field

try:
    from typing import Annotated  # py38 required Annotated
except ImportError:
    from typing_extensions import Annotated

PYDANTIC_EXTRAS = os.getenv("IPFABRIC_PYDANTIC_EXTRAS", "allow")


class Number(BaseModel, extra=PYDANTIC_EXTRAS):
    min: int
    max: int


class Numbers(BaseModel, extra=PYDANTIC_EXTRAS):
    """Possibly only for Internal Testing"""

    numbers: List[Number]


class Transport(BaseModel, extra=PYDANTIC_EXTRAS):
    src: Union[List[str], str]
    dst: Union[List[str], Numbers, str]


class TCP(Transport, BaseModel, extra=PYDANTIC_EXTRAS):
    flags: List[str]
    type: Literal["tcp"]


class UDP(Transport, BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["udp"]


class ICMP(BaseModel, extra=PYDANTIC_EXTRAS):
    icmpCode: int
    icmpType: int
    type: Literal["icmp"]


class MPLS(BaseModel, extra=PYDANTIC_EXTRAS):
    stack: List[int]
    type: Literal["mpls"]


class Ethernet(BaseModel, extra=PYDANTIC_EXTRAS):
    src: Optional[str] = None
    dst: Optional[str] = None
    etherType: str
    type: Literal["ethernet"]
    vlan: Optional[int] = None


class ESP(BaseModel, extra=PYDANTIC_EXTRAS):
    payload: str
    nextHeader: str
    type: Literal["esp"]


class IP(BaseModel, extra=PYDANTIC_EXTRAS):
    src: List[str]
    dst: List[str]
    fragmentOffset: int = Field(alias="fragment offset")
    protocol: str
    ttl: int
    type: Literal["ip"]


class VXLAN(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["vxlan"]
    vni: int
    groupPolicyId: Optional[int] = None
    policyApplied: Optional[bool] = None


class CAPWAP(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["capwap"]


class GRE(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["gre"]
    protoType: Optional[str] = None


class FabricPath(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["fp"]
    dstSubswitchId: int
    dstSwitchId: int
    etherType: str
    srcSwitchId: int
    ttl: int


PROTOCOLS = Annotated[
    Union[ICMP, UDP, TCP, Ethernet, IP, MPLS, ESP, VXLAN, CAPWAP, GRE, FabricPath], Field(discriminator="type")
]



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Mpls(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def ldp_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/mpls/ldp/neighbors", sn=self.sn)

    @property
    def ldp_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/mpls/ldp/interfaces", sn=self.sn)

    @property
    def rsvp_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/mpls/rsvp/neighbors", sn=self.sn)

    @property
    def rsvp_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/mpls/rsvp/interfaces", sn=self.sn)

    @property
    def forwarding(self):
        return models.Table(client=self.client, endpoint="tables/mpls/forwarding", sn=self.sn)

    @property
    def l3vpn_pe_routers(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l3-vpn/pe-routers", sn=self.sn)

    @property
    def l3vpn_pe_vrfs(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l3-vpn/pe-vrfs", sn=self.sn)

    @property
    def l3vpn_vrf_targets(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l3-vpn/vrf-targets", sn=self.sn)

    @property
    def l3vpn_pe_routes(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l3-vpn/pe-routes", sn=self.sn)

    @property
    def l2vpn_point_to_point_vpws(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l2-vpn/point-to-point-vpws", sn=self.sn)

    @property
    def l2vpn_point_to_multipoint(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l2-vpn/point-to-multipoint", sn=self.sn)

    @property
    def l2vpn_circuit_cross_connect(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l2-vpn/circuit-cross-connect", sn=self.sn)

    @property
    def l2vpn_pseudowires(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l2-vpn/pseudowires", sn=self.sn)



import logging
from typing import Any, Union, List

from pydantic.dataclasses import dataclass

from ipfabric.tools.shared import convert_timestamp, date_parser

logger = logging.getLogger("ipfabric")

COLUMNS = ["id", "sn", "hostname", "loginIp", "loginType", "ts", "username", "usernameNotes"]


@dataclass
class DiscoveryHistory:
    ipf: Any

    def get_all_history(
        self, columns: list = None, sort: dict = None, filters: dict = None, ts_format: str = "utc"
    ) -> list:
        """returns information about the discovery process

        Args:

        columns: columns for discovery-history table
        sort: Default timestamp descending
        filters: filter for discovery-history table
        ts_format: Valid formats ['utc', 'datetime', 'int']; datetime will return python datetime object

        Returns:
            List[Dict]
        """
        columns = columns or COLUMNS
        if ts_format not in ["utc", "datetime", "int"] and "ts" in columns:
            raise SyntaxError(f"{ts_format} is not a valid format; please use one of ['utc', 'datetime', 'int'].")
        sort = sort or {"order": "desc", "column": "ts"}
        history = self.ipf.fetch_all(
            "tables/inventory/discovery-history", columns=columns, filters=filters, sort=sort, snapshot=False
        )
        if ts_format == "datetime" and "ts" in columns:
            [h.update({"ts": convert_timestamp(h["ts"])}) for h in history]
        elif ts_format == "utc" and "ts" in columns:
            [h.update({"ts": convert_timestamp(h["ts"], ts_format="ziso")}) for h in history]
        return history

    def get_snapshot_history(
        self, snapshot_id: str = None, columns: list = None, sort: dict = None, ts_format: str = "utc"
    ) -> tuple:
        """returns information about a specific snapshot

        Args:
            snapshot_id: Specific ID or defaults to class initialized ID
            columns: Default All
            sort: Default timestamp descending
            ts_format: Valid formats ['utc', 'datetime', 'int']; datetime will return python datetime object

        Returns:
            First list has history of devices in the snapshot, second list has device that
            did not have history usually because it is an AP or API connection
        """
        snapshot = self.ipf.snapshots[snapshot_id or self.ipf.snapshot_id]
        hist = {
            h["sn"]: h
            for h in self.get_history_date(
                daterange=int(snapshot.start.timestamp()), columns=columns, sort=sort, ts_format=ts_format
            )
        }
        history, no_history = list(), list()
        for device in self.ipf.inventory.devices.all(columns=["sn", "hostname", "loginIp", "loginType"]):
            if device["sn"] in hist:
                history.append(hist[device["sn"]])
            else:
                no_history.append(device)
        return history, no_history

    def get_history_date(
        self, daterange: Union[tuple, str, int], columns: list = None, sort: dict = None, ts_format: str = "utc"
    ) -> list:
        """get a full history of a snapshot

        Args:
            daterange: Date can be string or int in seconds "11/22/ 1:30" or 1637629200
                        This will filter from that date to now
                        Or can be a tuple of a date range to filter history in that range.
            columns: Default All
            sort: Default timestamp descending
            ts_format: Valid formats ['utc', 'datetime', 'int']; datetime will return python datetime object

        Returns: First list has history of devices in the snapshot, second list has device that
                did not have history usually because it is an AP or API connection
        """
        if isinstance(daterange, tuple):
            filters = {
                "and": [
                    {"ts": ["gte", int(date_parser(daterange[0]).timestamp() * 1000)]},
                    {"ts": ["lte", int(date_parser(daterange[1]).timestamp() * 1000)]},
                ]
            }

        else:
            filters = {"ts": ["gte", int(date_parser(daterange).timestamp() * 1000)]}
        return self.get_all_history(columns=columns, sort=sort, filters=filters, ts_format=ts_format)

    def delete_history(self, history_id: Union[List[str], str]):
        if isinstance(history_id, str):
            history_id = [history_id]

        resp = self.ipf.request("DELETE", "discovery/history", json=history_id)
        resp.raise_for_status()
        logger.debug("Deleted the following discovery history IDs:")
        logger.debug(f"{history_id}")
        return True

    def delete_history_prior_to_ts(self, timestamp: Union[int, str]) -> list:
        """removes history from IPF based on the date passed to timestamp

        Args:
            timestamp: Can be string or int in seconds "11/22/ 1:30" or 1637629200
        Returns:
            List of history
        """
        filters = {"ts": ["lte", int(date_parser(timestamp).timestamp() * 1000)]}
        history = self.get_all_history(filters=filters)
        if not history:
            logger.debug(f"Did not locate any history prior to {date_parser(timestamp).isoformat()}")
            return history
        history_ids = [h["id"] for h in history]
        self.delete_history(history_ids)
        logger.debug("Deleted the following discovery history entries:")
        [logger.debug(f"{h}") for h in history]
        return history



import json
import re
from typing import List, Union, Literal

from pydantic import field_validator, BaseModel, Field, AnyHttpUrl
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

AWS_REGIONS = [
    "af-south-1",
    "ap-east-1",
    "ap-northeast-1",
    "ap-northeast-2",
    "ap-northeast-3",
    "ap-south-1",
    "ap-south-2",
    "ap-southeast-1",
    "ap-southeast-2",
    "ap-southeast-3",
    "ap-southeast-4",
    "ca-central-1",
    "eu-central-1",
    "eu-central-2",
    "eu-north-1",
    "eu-south-1",
    "eu-south-2",
    "eu-west-1",
    "eu-west-2",
    "eu-west-3",
    "il-central-1",
    "me-central-1",
    "me-south-1",
    "sa-east-1",
    "us-east-1",
    "us-east-2",
    "us-gov-east-1",
    "us-gov-west-1",
    "us-west-1",
    "us-west-2",
]

SLUG = r"^[a-zA-Z0-9_\-.#:]*$"

URL = Annotated[AnyHttpUrl, AfterValidator(lambda x: str(x))]


class DefaultRateLimiterSettings(BaseModel):
    maxConcurrentRequests: int = 100000
    maxCapacity: int = 300
    refillRate: int = 10
    refillRateIntervalMs: int = 1000


class VendorAPI(BaseModel):
    """
    base Vendor API config
    """

    slug: str
    comment: str = ""
    isEnabled: bool = True

    @field_validator("slug")
    @classmethod
    def check_slug(cls, slug):
        if not re.match(SLUG, slug):
            raise ValueError(f"{slug} is not a valid slug and must match regex {SLUG}")
        return slug


class SystemProxy(BaseModel):
    """
    support for Proxy Servers when utilizing Vendor APIs
    """

    respectSystemProxyConfiguration: bool = True


class RejectUnauthorized(SystemProxy, BaseModel):
    """
    support for credentials when utilizing Vendor APIs
    """

    rejectUnauthorized: bool = True


class UserAuthBaseUrl(BaseModel):
    """
    support for authentication when utilizing Vendor APIs
    """

    username: str
    password: str
    baseUrl: URL


class AssumeRole(BaseModel):
    role: str


class AWS(VendorAPI, SystemProxy, DefaultRateLimiterSettings, BaseModel):
    """
    AWS vendor api support
    """

    apiKey: str
    apiSecret: str
    regions: list
    assumeRoles: List[Union[str, dict, AssumeRole]] = Field(default_factory=list)
    type: Literal["aws-ec2"] = "aws-ec2"
    maxCapacity: int = 50

    @field_validator("regions")
    @classmethod
    def check_region(cls, regions):
        for r in regions:
            if r.lower() not in AWS_REGIONS:
                raise ValueError(f"{r} is not a valid AWS Region")
        return [r.lower() for r in regions]

    @field_validator("assumeRoles")
    @classmethod
    def check_roles(cls, roles):
        validated_roles = list()
        for role in roles:
            if isinstance(role, str):
                validated_roles.append(AssumeRole(role=role))
            elif isinstance(role, dict):
                if "role" in role:
                    validated_roles.append(AssumeRole(**role))
                else:
                    raise SyntaxError(f'Role {role} not in \'{{"role": "<arn:aws:iam::*****:role/*****>"}}\' format.')
            elif isinstance(role, AssumeRole):
                validated_roles.append(role)
        return validated_roles


class Azure(VendorAPI, SystemProxy, DefaultRateLimiterSettings, BaseModel):
    """
    Azure vendor api support
    """

    clientId: str
    clientSecret: str
    subscriptionId: str
    tenantId: str
    type: Literal["azure"] = "azure"
    maxConcurrentRequests: int = 1000


class CheckPointApiKey(VendorAPI, RejectUnauthorized, DefaultRateLimiterSettings, BaseModel):
    """
    Checkpoint vendor api support
    """

    apiKey: str
    baseUrl: URL
    domains: List[str] = Field(default_factory=list)
    type: Literal["checkpoint-mgmt-api"] = "checkpoint-mgmt-api"
    maxConcurrentRequests: int = 1000


class CheckPointUserAuth(VendorAPI, RejectUnauthorized, UserAuthBaseUrl, DefaultRateLimiterSettings, BaseModel):
    """
    checkpoint authentication vendor api support
    """

    domains: List[str] = Field(default_factory=list)
    type: Literal["checkpoint-mgmt-api"] = "checkpoint-mgmt-api"
    maxConcurrentRequests: int = 1000


class CiscoAPIC(VendorAPI, RejectUnauthorized, UserAuthBaseUrl, DefaultRateLimiterSettings, BaseModel):
    """
    Cisco APIC vendor api support
    """

    type: Literal["ciscoapic"] = "ciscoapic"
    maxConcurrentRequests: int = 10


class CiscoFMC(VendorAPI, RejectUnauthorized, UserAuthBaseUrl, DefaultRateLimiterSettings, BaseModel):
    """
    Cisco FMC vendor api support
    """

    type: Literal["ciscofmc"] = "ciscofmc"
    maxConcurrentRequests: int = 9
    maxCapacity: int = 110
    refillRate: int = 110
    refillRateIntervalMs: int = 60000


class ForcePoint(VendorAPI, RejectUnauthorized, DefaultRateLimiterSettings, BaseModel):
    """
    ForcePoint vendor api support
    """

    authenticationKey: str
    baseUrl: URL
    type: Literal["forcepoint"] = "forcepoint"
    maxConcurrentRequests: int = 10


class GCPCredentials(BaseModel):
    """
    GCP Credentials
    """

    auth_provider_x509_cert_url: str
    auth_uri: str
    client_email: str
    client_id: str
    client_x509_cert_url: str
    private_key: str
    private_key_id: str
    project_id: str
    token_uri: str
    type: str
    universe_domain: str


class GCP(VendorAPI, RejectUnauthorized, BaseModel):
    """
    GCP vendor api support

    Args:
        credentialsJson: JSON File, Dictionary, or String of GCP Credentials
    """

    credentialsJson: Union[GCPCredentials, dict, str]
    type: Literal["gcp"] = "gcp"

    @field_validator("credentialsJson")
    @classmethod
    def validate_credentials(cls, credentials):
        if isinstance(credentials, dict):
            return GCPCredentials(**credentials)
        elif isinstance(credentials, str):
            try:
                with open(credentials, "r") as f:
                    return GCPCredentials(**json.load(f))
            except FileNotFoundError:
                try:
                    return GCPCredentials(**json.loads(credentials))
                except json.JSONDecodeError:
                    raise ValueError("credentialsJson must be a valid JSON file, JSON string or Python dictionary.")
        else:
            return credentials


class JuniperMist(VendorAPI, RejectUnauthorized, DefaultRateLimiterSettings, BaseModel):
    """
    Juniper Mist vendor api support
    """

    apiToken: str
    apiVer: Literal["v1"] = "v1"
    type: Literal["juniper-mist"] = "juniper-mist"
    baseUrl: URL = "https://api.mist.com"
    maxConcurrentRequests: int = 250


class Merakiv1(VendorAPI, RejectUnauthorized, DefaultRateLimiterSettings, BaseModel):
    """
    Meraki v1 vendor api support
    """

    apiKey: str
    baseUrl: URL
    organizations: List[str] = Field(default_factory=list)
    apiVer: Literal["v1"] = "v1"
    type: Literal["meraki-v0"] = "meraki-v0"
    maxCapacity: int = 10


class Prisma(VendorAPI, RejectUnauthorized, DefaultRateLimiterSettings, BaseModel):
    """
    Prisma vendor api support
    """

    username: str
    password: str
    tsgid: str
    type: Literal["prisma"] = "prisma"
    maxConcurrentRequests: int = 4


class RuckusVirtualSmartZone(VendorAPI, RejectUnauthorized, UserAuthBaseUrl, DefaultRateLimiterSettings, BaseModel):
    """
    Ruckus Virtual SmartZone vendor api support
    """

    apiVer: Literal["v9_1"] = "v9_1"
    type: Literal["ruckus-vsz"] = "ruckus-vsz"
    maxConcurrentRequests: int = 10


class SilverPeak(VendorAPI, RejectUnauthorized, UserAuthBaseUrl, DefaultRateLimiterSettings, BaseModel):
    """
    SilverPeak vendor api support
    """

    loginType: str = "Local"
    type: Literal["silverpeak"] = "silverpeak"
    maxConcurrentRequests: int = 5

    @field_validator("loginType")
    @classmethod
    def check_login(cls, login_type):
        if login_type not in ["Local", "RADIUS", "TACACS+"]:
            raise ValueError(f"{login_type} is not a valid login type must be in ['Local', 'RADIUS', 'TACACS+']")
        return login_type


class Versa(VendorAPI, RejectUnauthorized, UserAuthBaseUrl, DefaultRateLimiterSettings, BaseModel):
    """
    Versa vendor api support
    """

    paginationLimit: int = 1000
    type: Literal["versa"] = "versa"
    maxConcurrentRequests: int = 10
    maxCapacity: int = 100
    refillRate: int = 2


class Viptela(VendorAPI, RejectUnauthorized, UserAuthBaseUrl, DefaultRateLimiterSettings, BaseModel):
    """
    Viptela vendor api support
    """

    type: Literal["viptela"] = "viptela"
    maxConcurrentRequests: int = 25


class NSXT(VendorAPI, RejectUnauthorized, UserAuthBaseUrl, BaseModel):
    """
    NSXT vendor api support
    """

    type: Literal["nsxT"] = "nsxT"
    maxConcurrentRequests: int = 40
    maxCapacity: int = 100
    refillRate: int = 50



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Wireless(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def controllers(self):
        return models.Table(client=self.client, endpoint="tables/wireless/controllers", sn=self.sn)

    @property
    def access_points(self):
        return models.Table(client=self.client, endpoint="tables/wireless/access-points", sn=self.sn)

    @property
    def radios_detail(self):
        return models.Table(client=self.client, endpoint="tables/wireless/radio", sn=self.sn)

    @property
    def radios_ssid_summary(self):
        return models.Table(client=self.client, endpoint="tables/wireless/ssid-summary", sn=self.sn)

    @property
    def clients(self):
        return models.Table(client=self.client, endpoint="tables/wireless/clients", sn=self.sn)



import importlib.resources
import json
from typing import List

from pydantic import BaseModel


class Label(BaseModel):
    name: str
    center: bool = False

    def __str__(self):
        return self.name.lower()

    def __eq__(self, other):
        return self.name.lower() == other.lower()


class Protocol(BaseModel):
    name: str
    labels: List[Label]

    def __eq__(self, other):
        return self.name == other.lower()


PROTOCOL = Label(name="protocol", center=True)
INT_NAME = Label(name="intName", center=False)
IP_ADDR = Label(name="ipAddress", center=False)
SUBNET = Label(name="subnet", center=True)
AS = Label(name="as", center=False)
VALID_PROTOCOL_LABELS = dict(
    dgw=Protocol(name="dgw", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    dgwIpv6=Protocol(name="dgwIpv6", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    ebgp=Protocol(name="ebgp", labels=[PROTOCOL, IP_ADDR, AS, Label(name="prefix", center=False)]),
    eigrp=Protocol(name="eigrp", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET, AS]),
    fex=Protocol(name="fex", labels=[PROTOCOL, INT_NAME]),
    ibgp=Protocol(name="ibgp", labels=[PROTOCOL, IP_ADDR, AS, Label(name="prefix", center=False)]),
    isis=Protocol(name="isis", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    ldp=Protocol(name="ldp", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    ospf=Protocol(
        name="ospf",
        labels=[
            PROTOCOL,
            INT_NAME,
            IP_ADDR,
            SUBNET,
            Label(name="area", center=False),
            Label(name="cost", center=False),
        ],
    ),
    ospfv3=Protocol(
        name="ospfv3",
        labels=[
            PROTOCOL,
            INT_NAME,
            IP_ADDR,
            SUBNET,
            Label(name="area", center=False),
            Label(name="cost", center=False),
        ],
    ),
    pim=Protocol(name="pim", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    rib=Protocol(name="rib", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    ribIpv6=Protocol(name="ribIpv6", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    rip=Protocol(name="rip", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    rsvp=Protocol(name="rsvp", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    stp=Protocol(name="stp", labels=[PROTOCOL, INT_NAME, Label(name="media", center=False)]),
    vxlan=Protocol(name="vxlan", labels=[PROTOCOL, INT_NAME, IP_ADDR, SUBNET]),
    wired=Protocol(name="wired", labels=[PROTOCOL, INT_NAME]),
    wireless=Protocol(name="wireless", labels=[PROTOCOL, INT_NAME]),
    xdp=Protocol(name="xdp", labels=[PROTOCOL, INT_NAME]),
)

VALID_NET_PROTOCOLS = {p.lower(): p for p in VALID_PROTOCOL_LABELS.keys()}

VALID_DEV_TYPES = [
    "aciLeaf",
    "aciSpine",
    "ap",
    "cloud",
    "cloudInstance",
    "cloudInternetGw",
    "cloudLoadBalancer",
    "cloudNatGw",
    "cloudRouter",
    "cloudTransitHub",
    "cloudVpnGw",
    "fex",
    "fw",
    "host",
    "l3switch",
    "lb",
    "nx7000",
    "phone",
    "router",
    "securityManagement",
    "switch",
    "transit",
    "unknown",
    "vgw",
    "waas",
    "wlc",
]

VALID_PATH_PROTOCOLS = ["capwap", "esp", "ethernet", "ethernetVlan", "gre", "ip", "loop", "mpls", "tcp", "udp", "vxlan"]

DEFAULT_NETWORK = json.loads(
    importlib.resources.read_text("ipfabric.diagrams.input_models.factory_defaults", "networkSettings.json")
)
DEFAULT_PATHLOOKUP = json.loads(
    importlib.resources.read_text("ipfabric.diagrams.input_models.factory_defaults", "pathLookupSettings.json")
)

VALID_LAYOUTS = ["circular", "downwardTree", "hierarchical", "radial", "universal", "upwardTree", "userDefined"]



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Addressing(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def arp_table(self):
        return models.Table(client=self.client, endpoint="tables/addressing/arp", sn=self.sn)

    @property
    def mac_table(self):
        return models.Table(client=self.client, endpoint="tables/addressing/mac", sn=self.sn)

    @property
    def managed_ip_ipv4(self):
        return models.Table(client=self.client, endpoint="tables/addressing/managed-devs", sn=self.sn)

    @property
    def managed_ip_ipv6(self):
        return models.Table(client=self.client, endpoint="tables/addressing/ipv6-managed-devs", sn=self.sn)

    @property
    def managed_duplicate_ip(self):
        return models.Table(client=self.client, endpoint="tables/addressing/duplicate-ip", sn=self.sn)

    @property
    def nat44(self):
        return models.Table(client=self.client, endpoint="tables/security/nat44", sn=self.sn)

    @property
    def ipv6_neighbor_discovery(self):
        return models.Table(client=self.client, endpoint="tables/addressing/ipv6-neighbors", sn=self.sn)



import logging
import re
from typing import Optional, List, Any

from pydantic import Field
from pydantic.dataclasses import dataclass

logger = logging.getLogger("ipfabric")

ATTR_REGEX = re.compile(r"^[a-zA-Z]\w*[a-zA-Z0-9]+$")


@dataclass
class Attributes:
    """
    Class to retrieve Global and Local (Snapshot specific) Attributes.  You must specify snapshot_id to use local.
    """

    client: Any = Field(description="IPFClient")
    snapshot_id: Optional[str] = Field(default=None, description="Snapshot ID to switch to Local Attributes")

    def __post_init__(self):
        if isinstance(self.snapshot_id, str):
            self.snapshot_id = self.client.get_snapshot(self.snapshot_id).snapshot_id

    @property
    def endpoint(self):
        return "attributes/local" if self.snapshot_id else "attributes/global"

    @property
    def post_endpoint(self):
        return "tables/snapshot-attributes" if self.snapshot_id else "tables/global-attributes"

    @staticmethod
    def check_attribute_name(attributes: set):
        invalid = list()
        for attribute in attributes:
            if not ATTR_REGEX.match(attribute):
                invalid.append(attribute)
        if invalid:
            raise NameError(
                f"The following Attribute Names are invalid and do match regex rule "
                f'"^[a-zA-Z][a-zA-Z0-9_]*[a-zA-Z0-9]+$":\n{invalid}'
            )
        return True

    def all(
        self,
        columns: list = None,
        filters: Optional[dict] = None,
        sort: Optional[dict] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            columns: Optional columns to return, default is all
            filters: Optional filters
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}

        Returns:
            list: List of Dictionaries
        """
        return self.client.fetch_all(
            self.post_endpoint,
            columns=columns,
            filters=filters,
            sort=sort,
            snapshot=True if self.snapshot_id else False,
            snapshot_id=self.snapshot_id,
        )

    def set_attribute_by_sn(self, serial_number, name, value):
        """
        Set a single Attribute by serial number.
        If Global will Error if already set.
        If Local will not error and either Add or Update
        :param serial_number: str: IP Fabric Unique Serial Number
        :param name: str: Attribute name (case sensitive)
        :param value: str: Attribute value (case sensitive)
        :return:
        """
        self.check_attribute_name({name})
        attribute = dict(name=name, sn=serial_number, value=value)
        if self.snapshot_id:
            return self.set_attributes_by_sn([attribute])
        resp = self.client.post(self.endpoint, json=attribute)
        resp.raise_for_status()
        return resp.json()

    def set_attributes_by_sn(self, attributes: List[dict]):
        """
        Sets a list of Attributes for devices based on serial numbers.
        Will Add or Update Attributes.
        :param attributes: list: [{'sn': 'IPF SERIAL NUMBER', 'name': 'attributeName', 'value': 'SITE NAME'}]
        :return:
        """
        self.check_attribute_name({v["name"] for v in attributes})
        payload = dict(attributes=attributes)
        if self.snapshot_id:
            payload["snapshot"] = self.snapshot_id
        resp = self.client.put(self.endpoint, json=payload)
        resp.raise_for_status()
        return resp.json()

    def set_site_by_sn(self, serial_number, site_name):
        """
        Set a single site by serial number
        If Global will Error if already set.
        If Local will not error and either Add or Update
        :param serial_number: str: IP Fabric Unique Serial Number
        :param site_name: str: Site name for device.
        :return:
        """
        return self.set_attribute_by_sn(serial_number, "siteName", site_name)

    def set_sites_by_sn(self, sites: List[dict]):
        """Sets a list of sites for devices based on serial numbers.

        Args:
            sites: [{'sn': 'IPF SERIAL NUMBER', 'value': 'SITE NAME'}]
        """
        [a.update(dict(name="siteName")) for a in sites]
        return self.set_attributes_by_sn(sites)

    def delete_attribute_by_sn(self, *serial_numbers):
        """Deletes attributes by Unique IP Fabric Serial Number(s)

        Args:
            serial_numbers: Serial Numbers
        """
        payload = dict(attributes=dict(sn=[str(i) for i in serial_numbers]))
        if self.snapshot_id:
            payload["snapshot"] = self.snapshot_id
        resp = self.client.request("DELETE", self.endpoint, json=payload)
        resp.raise_for_status()
        return True

    def delete_attribute_by_id(self, *attribute_ids):
        """Deletes attributes by Attribute ID(s)

        Args:
            attribute_ids: Attribute IDs
        """
        payload = dict(attributes=dict(id=[str(i) for i in attribute_ids]))
        if self.snapshot_id:
            payload["snapshot"] = self.snapshot_id
        resp = self.client.request("DELETE", self.endpoint, json=payload)
        resp.raise_for_status()
        return True

    def delete_attribute(self, *attributes):
        """
        Deletes attributes by Attribute
        :param attributes: dict: Attribute dictionaries
        :return:
        """
        return self.delete_attribute_by_id(*[str(i["id"]) for i in attributes])

    def update_local_attr_from_global(self) -> bool:
        """Currently can only update devices that are in the snapshot."""
        if not self.snapshot_id:
            raise ImportError("Please initialize Attributes class with a snapshot_id.")
        g_attrs = self.client.fetch_all("tables/global-attributes", columns=["name", "value", "sn"], snapshot=False)
        if not g_attrs:
            raise ValueError("No Global Attributes found.")

        local_attrs = self.all()
        if local_attrs:
            self.delete_attribute(*local_attrs)

        if self.client.api_version < "v6.7":  # TODO: Remove in v6.9 or v7.0
            local_sn = {
                i["sn"] for i in self.client.inventory.devices.all(columns=["sn"], snapshot_id=self.snapshot_id)
            }
            g_attrs = [g for g in g_attrs if g["sn"] in local_sn]
        self.set_attributes_by_sn(g_attrs)
        return True



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Security(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def acl(self):
        return models.Table(client=self.client, endpoint="tables/security/acl", sn=self.sn)

    @property
    def acl_interface(self):
        return models.Table(client=self.client, endpoint="tables/security/acl/interfaces", sn=self.sn)

    @property
    def acl_global_policies(self):
        return models.Table(client=self.client, endpoint="tables/security/acl/global-policies", sn=self.sn)

    @property
    def dmvpn(self):
        return models.Table(client=self.client, endpoint="tables/security/dmvpn", sn=self.sn)

    @property
    def dhcp_snooping(self):
        return models.Table(client=self.client, endpoint="tables/security/dhcp/snooping", sn=self.sn)

    @property
    def dhcp_snooping_bindings(self):
        return models.Table(client=self.client, endpoint="tables/security/dhcp/bindings", sn=self.sn)

    @property
    def ipsec_tunnels(self):
        return models.Table(client=self.client, endpoint="tables/security/ipsec/tunnels", sn=self.sn)

    @property
    def ipsec_gateways(self):
        return models.Table(client=self.client, endpoint="tables/security/ipsec/gateways", sn=self.sn)

    @property
    def secure_ports_devices(self):
        return models.Table(client=self.client, endpoint="tables/security/secure-ports/devices", sn=self.sn)

    @property
    def secure_ports_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/security/secure-ports/interfaces", sn=self.sn)

    @property
    def secure_ports_users(self):
        return models.Table(client=self.client, endpoint="tables/security/secure-ports/users", sn=self.sn)

    @property
    def zone_firewall_policies(self):
        return models.Table(client=self.client, endpoint="tables/security/zone-firewall/policies", sn=self.sn)

    @property
    def zone_firewall_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/security/zone-firewall/interfaces", sn=self.sn)



import logging
from typing import Any

from pydantic import BaseModel
from pydantic.dataclasses import dataclass

logger = logging.getLogger("ipfabric")


class Networks(BaseModel):
    exclude: list[str]
    include: list[str]


@dataclass
class Discovery:
    client: Any

    @property
    def networks(self):
        res = self.client.get("settings")
        res.raise_for_status()
        return Networks(**res.json()["networks"])

    def update_discovery_networks(self, subnets: list, include: bool = False):
        payload = dict()
        payload["networks"] = dict()
        if include:
            payload["networks"]["include"] = subnets
            payload["networks"]["exclude"] = self.networks.exclude
        else:
            payload["networks"]["exclude"] = subnets
            payload["networks"]["include"] = self.networks.include
        res = self.client.patch("settings", json=payload)
        res.raise_for_status()
        return Networks(**res.json()["networks"])



from collections import defaultdict
from copy import deepcopy
from typing import Optional

from httpx._types import ProxiesTypes
from pydantic import BaseModel

from ipfabric.tools.nist import NIST, CVEs


class Version(BaseModel):
    vendor: str
    family: Optional[str] = None
    version: str
    cves: CVEs
    hostname: Optional[str] = None
    site: Optional[str] = None


class Vulnerabilities:
    def __init__(self, ipf, nvd_api_key: str, timeout: int = 30, proxies: Optional[ProxiesTypes] = None):
        self.ipf = ipf
        self.nist = NIST(nvd_api_key=nvd_api_key, timeout=timeout, proxies=proxies)

    def __del__(self):
        try:
            self.nist.close()
        except AttributeError:
            return

    def _check_versions(self, versions):
        cves = list()
        for v in versions:
            cve = self.nist.check_cve(v["vendor"], v["family"], v["version"])
            cves.append(Version(vendor=v["vendor"], family=v["family"], version=v["version"], cves=cve))
        return cves

    def _check_devices(self, devices):
        versions = [
            {"vendor": v[0], "family": v[1], "version": v[2]}
            for v in {(d["vendor"], d["family"], d["version"]) for d in devices}
        ]
        cve_dict = defaultdict(dict)
        for c in self._check_versions(versions):
            if c.vendor not in cve_dict:
                cve_dict[c.vendor] = defaultdict(dict)
            cve_dict[c.vendor][c.family].update({c.version: c})
        cves = list()
        for d in devices:
            cve = deepcopy(cve_dict[d["vendor"]][d["family"]][d["version"]])
            cve.hostname = d["hostname"]
            cve.site = d["siteName"]
            cves.append(cve)
        return cves

    def check_versions(self, vendor=None):
        filters = {"vendor": ["like", vendor]} if vendor else None
        versions = self.ipf.fetch_all(
            "tables/management/osver-consistency",
            columns=["vendor", "family", "version"],
            filters=filters,
        )
        return self._check_versions(versions)

    def check_device(self, device):
        devices = self.ipf.inventory.devices.all(
            columns=["hostname", "siteName", "vendor", "family", "version"],
            filters={"hostname": ["like", device]},
        )
        return self._check_devices(devices)

    def check_site(self, site):
        devices = self.ipf.inventory.devices.all(
            columns=["hostname", "siteName", "vendor", "family", "version"],
            filters={"siteName": ["like", site]},
        )
        return self._check_devices(devices)



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Dhcp(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def relay_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/interfaces", sn=self.sn)

    @property
    def relay_interfaces_stats_received(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/interfaces-stats/received", sn=self.sn)

    @property
    def relay_interfaces_stats_relayed(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/interfaces-stats/relayed", sn=self.sn)

    @property
    def relay_interfaces_stats_sent(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/interfaces-stats/sent", sn=self.sn)

    @property
    def relay_global_stats_summary(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/global-stats/summary", sn=self.sn)

    @property
    def relay_global_stats_received(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/global-stats/received", sn=self.sn)

    @property
    def relay_global_stats_relayed(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/global-stats/relayed", sn=self.sn)

    @property
    def relay_global_stats_sent(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/global-stats/sent", sn=self.sn)

    @property
    def server_summary(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/summary", sn=self.sn)

    @property
    def server_pools(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/pools", sn=self.sn)

    @property
    def server_leases(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/leases", sn=self.sn)

    @property
    def server_excluded_ranges(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/excluded-ranges", sn=self.sn)

    @property
    def server_excluded_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/interfaces", sn=self.sn)



import os
import re
from typing import Optional, List, Union
from uuid import UUID

from pydantic import field_validator, BaseModel, Field
from pydantic_extra_types.color import ColorType, Color

from ipfabric.tools import VALID_REFS
from .constants import (
    VALID_DEV_TYPES,
    DEFAULT_NETWORK,
    DEFAULT_PATHLOOKUP,
    VALID_NET_PROTOCOLS,
    VALID_PROTOCOL_LABELS,
    VALID_PATH_PROTOCOLS,
)

PYDANTIC_EXTRAS = os.getenv("IPFABRIC_PYDANTIC_EXTRAS", "allow")


class Style(BaseModel, extra=PYDANTIC_EXTRAS):
    color: ColorType
    pattern: Optional[str] = "solid"
    thicknessThresholds: Optional[List[int]] = [2, 4, 8]

    @field_validator("pattern")
    @classmethod
    def _valid_patterns(cls, v):
        if v.lower() not in ["solid", "dashed", "dotted"]:
            raise ValueError(f'Pattern "{v}" not in ["solid", "dashed", "dotted"]')
        return v.lower()

    @field_validator("color")
    @classmethod
    def _parse_color(cls, v):
        return Color(v)

    def style_settings(self) -> dict:
        return dict(color=self.color.as_hex(), pattern=self.pattern, thicknessThresholds=self.thicknessThresholds)


class Setting(BaseModel, extra=PYDANTIC_EXTRAS):
    name: str
    visible: bool = True
    grouped: bool = True
    style: Style
    type: str

    def base_settings(self) -> dict:
        return dict(
            name=self.name,
            visible=self.visible,
            grouped=self.grouped,
            style=self.style.style_settings(),
            type=self.type,
        )


class EdgeSettings(Setting, BaseModel, extra=PYDANTIC_EXTRAS):
    labels: List[str] = ["protocols"]
    id: Optional[UUID] = None

    def settings(self) -> dict:
        base_settings = self.base_settings()
        base_settings["labels"] = self.labels
        return base_settings


class GroupSettings(Setting, BaseModel, extra=PYDANTIC_EXTRAS):
    label: str
    children: List[EdgeSettings]
    id: Optional[UUID] = None

    def settings(self) -> dict:
        base_settings = self.base_settings()
        base_settings.update(dict(children=[child.settings() for child in self.children], label=self.label))
        return base_settings


class PathLookup(BaseModel, extra=PYDANTIC_EXTRAS):
    ignoredTopics: Optional[List[str]] = Field(
        default_factory=list,
        description="List of topics to ignore.  Valid topics are in ['ACL', 'FORWARDING', 'ZONEFW'].",
    )
    colorDetectedLoops: Optional[bool] = True

    @field_validator("ignoredTopics")
    @classmethod
    def _valid_topics(cls, v):
        if v and not all(t in ["ACL", "FORWARDING", "ZONEFW"] for t in v):
            raise ValueError(f"Ignored Topics '{v}' must be None or in ['ACL', 'FORWARDING', 'ZONEFW'].")
        return v


class GraphSettings(BaseModel, extra=PYDANTIC_EXTRAS):
    edges: List[Union[GroupSettings, EdgeSettings]]
    hiddenDeviceTypes: Optional[List[str]] = Field(default_factory=list)
    pathLookup: Optional[PathLookup] = None

    @field_validator("hiddenDeviceTypes")
    @classmethod
    def _valid_dev_types(cls, v):
        if v and not all(d in VALID_DEV_TYPES for d in v):
            raise ValueError(f"Device Types '{v}' must be None or in {VALID_DEV_TYPES}.")
        return v


class NetworkSettings(GraphSettings):
    def __init__(self):
        edges = [GroupSettings(**edge) for edge in DEFAULT_NETWORK]
        super().__init__(edges=edges, hiddenDeviceTypes=["ap", "fex", "host", "phone"])

    @staticmethod
    def _update_edge(children: List[EdgeSettings], name: str, attribute: str, bool_value=False):
        for edge in children:
            if edge.name.lower() == name:
                setattr(edge, attribute, bool_value)
                return True
        return False

    def _update_group(self, name: str, attribute: str, group: bool = False, bool_value=False):
        for edge in self.edges:
            if group and isinstance(edge, GroupSettings) and edge.name.lower() == name:
                setattr(edge, attribute, bool_value)
                return True
            elif not group:
                if isinstance(edge, GroupSettings) and self._update_edge(edge.children, name, attribute, bool_value):
                    return self._update_group(edge.name.lower(), "grouped", True)
                elif isinstance(edge, EdgeSettings) and self._update_edge([edge], name, attribute, bool_value):
                    return True
        return False

    def hide_protocol(self, protocol_name: str, unhide: bool = False):
        if protocol_name.lower() in VALID_NET_PROTOCOLS:
            return self._update_group(
                VALID_NET_PROTOCOLS[protocol_name.lower()], attribute="visible", group=False, bool_value=unhide
            )
        else:
            raise KeyError(
                f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_NET_PROTOCOLS.values()}"
            )

    def hide_all_protocols(self):
        for protocol_name in VALID_NET_PROTOCOLS.values():
            self._update_group(protocol_name, attribute="visible", group=False)

    def ungroup_protocol(self, protocol_name: str):
        if protocol_name.lower() in VALID_NET_PROTOCOLS:
            return self._update_group(VALID_NET_PROTOCOLS[protocol_name.lower()], attribute="grouped", group=False)
        else:
            raise KeyError(
                f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_NET_PROTOCOLS.values()}"
            )

    def hide_group(self, group_name: str):
        group_names = [g.name.lower() for g in self.edges if isinstance(g, GroupSettings)]
        if group_name.lower() in group_names:
            return self._update_group(group_name.lower(), attribute="visible", group=True)
        else:
            raise KeyError(f"Group {group_name} does not exist.  Valid groups are {group_names}")

    def ungroup_group(self, group_name: str):
        group_names = [g.name.lower() for g in self.edges if isinstance(g, GroupSettings)]
        if group_name.lower() in group_names:
            return self._update_group(group_name.lower(), attribute="grouped", group=True)
        else:
            raise KeyError(f"Group {group_name} does not exist.  Valid groups are {group_names}")

    @staticmethod
    def _proto_label(edge: EdgeSettings, protocol_name: str, label_name: str):
        if edge.name.lower() == protocol_name:
            proto = next(x for x in VALID_PROTOCOL_LABELS[protocol_name].labels if x == label_name)
            if proto.center:
                edge.labels[0] = proto.name
            else:
                edge.labels[1] = proto.name
            return True
        return False

    def change_label(self, protocol_name: str, label_name: str):
        protocol_name, label_name = protocol_name.lower(), label_name.lower()
        if protocol_name not in VALID_NET_PROTOCOLS:
            raise KeyError(
                f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_NET_PROTOCOLS.values()}"
            )
        else:
            protocol_name = VALID_NET_PROTOCOLS[protocol_name]
        if label_name not in VALID_PROTOCOL_LABELS[protocol_name].labels:
            raise KeyError(
                f"Label {label_name} does not exist for protocol {protocol_name}.  "
                f"Valid labels for {protocol_name} are {VALID_PROTOCOL_LABELS[protocol_name].labels}"
            )
        for edge in self.edges:
            if isinstance(edge, GroupSettings):
                for child in edge.children:
                    if self._proto_label(child, protocol_name, label_name):
                        return True
            if self._proto_label(edge, protocol_name, label_name):
                return True
        return False

    def settings(self) -> dict:
        settings = dict(
            edges=[edge.settings() for edge in self.edges],
            hiddenDeviceTypes=self.hiddenDeviceTypes,
        )
        return settings


class PathLookupSettings(GraphSettings):
    def __init__(self):
        edges = [EdgeSettings(**edge) for edge in DEFAULT_PATHLOOKUP]
        super().__init__(edges=edges, pathLookup=PathLookup())

    @property
    def protocol_priority(self):
        return {edge.name.lower(): idx for idx, edge in enumerate(self.edges)}

    def increase_priority(self, protocol_name: str):
        if protocol_name.lower() not in VALID_PATH_PROTOCOLS:
            raise KeyError(f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_PATH_PROTOCOLS}")
        current = self.protocol_priority[protocol_name]
        if current != 0:
            self.edges[current], self.edges[current - 1] = self.edges[current - 1], self.edges[current]
        return True

    def decrease_priority(self, protocol_name: str):
        if protocol_name.lower() not in VALID_PATH_PROTOCOLS:
            raise KeyError(f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_PATH_PROTOCOLS}")
        current = self.protocol_priority[protocol_name]
        if current != len(self.edges) - 1:
            self.edges[current], self.edges[current + 1] = self.edges[current + 1], self.edges[current]
        return True

    def settings(self) -> dict:
        settings = dict(
            edges=[edge.settings() for edge in self.edges],
            pathLookup=vars(self.pathLookup),
        )
        return settings


class Overlay(BaseModel):
    """Set snapshotToCompare or intentRuleId, not both."""

    snapshotToCompare: Optional[Union[UUID, str]] = Field(None, description="Snapshot ID to compare.")
    intentRuleId: Optional[Union[int, str]] = Field(
        None,
        description="Intent Rule ID to overlay. Also valid: ['nonRedundantEdges', 'singlePointsOfFailure']",
    )

    @field_validator("snapshotToCompare")
    @classmethod
    def _valid_snapshot(cls, v):
        if v and v in VALID_REFS:
            return v
        elif v and isinstance(v, UUID):
            return str(v)
        elif re.match(r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$", v.lower()):
            return v.lower()
        raise ValueError(f'"{v}" is not a Snapshot ID or in {VALID_REFS}')

    @field_validator("intentRuleId")
    @classmethod
    def _valid_intentrule(cls, v):
        if v and v in ["nonRedundantEdges", "singlePointsOfFailure"]:
            return v
        try:
            return str(int(v))
        except ValueError:
            raise ValueError(f'"{v}" is not an Intent Rule ID or in ["nonRedundantEdges", "singlePointsOfFailure"]')

    @property
    def type(self):
        return "compare" if self.snapshotToCompare else "intent"

    def overlay(self) -> dict:
        if self.snapshotToCompare:
            return dict(type=self.type, snapshotToCompare=self.snapshotToCompare)
        else:
            return dict(type=self.type, intentRuleId=self.intentRuleId)



import json
import logging
from collections import OrderedDict
from datetime import datetime
from os import path
from typing import Any, Union

from pydantic.dataclasses import dataclass

logger = logging.getLogger("ipfabric")

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

INTENTS = "intents.json"
GROUPS = "groups.json"
DASHBOARD = "dashboard.json"

V4_4_PATH = files("ipfabric.tools.factory_defaults.v4.4")
V5_0_PATH = files("ipfabric.tools.factory_defaults.v5.0")
V6_0_PATH = files("ipfabric.tools.factory_defaults.v6.0")
V6_3_PATH = files("ipfabric.tools.factory_defaults.v6.3")

DEFAULT_SETUP = {
    "v4.4": {
        "intents": json.loads(V4_4_PATH.joinpath(INTENTS).read_text()),
        "groups": json.loads(V4_4_PATH.joinpath(GROUPS).read_text()),
        "dashboard": json.loads(V4_4_PATH.joinpath(DASHBOARD).read_text(), object_pairs_hook=OrderedDict),
    },
    "v5.0": {
        "intents": json.loads(V5_0_PATH.joinpath(INTENTS).read_text()),
        "groups": json.loads(V5_0_PATH.joinpath(GROUPS).read_text()),
        "dashboard": json.loads(V5_0_PATH.joinpath(DASHBOARD).read_text(), object_pairs_hook=OrderedDict),
    },
    "v6.0": {
        "intents": json.loads(V6_0_PATH.joinpath(INTENTS).read_text()),
        "groups": json.loads(V6_0_PATH.joinpath(GROUPS).read_text()),
        "dashboard": json.loads(V6_0_PATH.joinpath(DASHBOARD).read_text(), object_pairs_hook=OrderedDict),
    },
    "v6.3": {
        "intents": json.loads(V6_3_PATH.joinpath(INTENTS).read_text()),
        "groups": json.loads(V6_3_PATH.joinpath(GROUPS).read_text()),
        "dashboard": json.loads(V6_3_PATH.joinpath(DASHBOARD).read_text(), object_pairs_hook=OrderedDict),
    },
}


@dataclass
class RestoreIntents:
    ipf: Any

    def restore_from_file(self, intents_file: str, groups_file: str, dashboard_file: str):
        if not all(path.exists(f) for f in [intents_file, groups_file, dashboard_file]):
            logger.critical("File(s) do not exist.")
            exit()
        with open(intents_file, "r") as file:
            intents = json.load(file)
        with open(groups_file, "r") as file:
            groups = json.load(file)
        with open(dashboard_file, "r") as file:
            dashboard = json.load(file)
        self._restore(intents, groups, dashboard)

    def restore_from_dictionary(self, intents: list, groups: list, dashboard: Union[dict, OrderedDict]):
        if not all(isinstance(f, list) for f in [intents, groups]) and not isinstance(dashboard, (dict, OrderedDict)):
            logger.critical("Intents and Groups must be Lists and Dashboard a Dictionary or OrderedDict.")
            exit()
        self._restore(intents, groups, dashboard)

    def restore_default(self, version=None):
        v = version.lower() or "v" + str(self.ipf.os_version)[0:3]
        if v not in DEFAULT_SETUP:
            logger.critical(f"Version {v} not in Defaults {list(DEFAULT_SETUP)}.")
            exit()
        self._restore(DEFAULT_SETUP[v]["intents"], DEFAULT_SETUP[v]["groups"], DEFAULT_SETUP[v]["dashboard"])

    @staticmethod
    def _save_to_file(intents, groups, dashboard):
        tstamp = datetime.now().strftime("%Y%m%d-%H%M%S")

        def save(name, data):
            filename = name + "_" + tstamp + ".json"
            logger.info(f"Saving {name} to {filename}.")
            with open(filename, "w") as file:
                json.dump(data, file)
            logger.info(f"Saved {name} to {filename}.")

        save("intents", intents)
        save("groups", groups)
        save("dashboard", dashboard)

    def _restore(self, intents, groups, dashboard):
        i = input("\nDO YOU ACCEPT OVERRIDING ALL INTENTS, INTENT GROUPS, DASHBOARD SETTINGS? (y/n): ").strip().lower()
        if i and "y" not in i[0]:
            logger.warning("Exiting Script before making changes.")
            exit()

        old_intents = self.ipf.get("reports").json()
        old_intent_groups = self.ipf.get("reports/groups").json()
        old_dashboard = json.loads(self.ipf.get("settings/dashboard").text, object_pairs_hook=OrderedDict)
        self._save_to_file(old_intents, old_intent_groups, old_dashboard)

        logger.info("Deleting All Intents.")
        for i in old_intents:
            res = self.ipf.delete(f"reports/{i['id']}")
            res.raise_for_status()
        logger.info("Deleted All Intents.")

        logger.info("Deleting All Intent Groups.")
        for g in old_intent_groups:
            res = self.ipf.delete(f"reports/groups/{g['id']}")
            res.raise_for_status()
        logger.info("Deleted All Intent Groups.")

        intent_mapping = {i["id"]: None for i in intents}
        logger.info("Creating Intent Verification Rules.")
        for i in intents:
            i["groups"] = []
            i.pop("custom", None)
            src_id = i.pop("id", None)
            res = self.ipf.post("reports", json=i)
            res.raise_for_status()
            intent_mapping[src_id] = res.json()["id"]
        logger.info("Created Intent Verification Rules.")

        group_mapping = {g["id"]: None for g in groups}
        logger.info("Creating Intent Groups.")
        for g in groups:
            for c in g["children"]:
                c["id"] = intent_mapping[c["id"]]
            src_id = g.pop("id", None)
            g.pop("custom", None)
            res = self.ipf.post("reports/groups", json=g)
            res.raise_for_status()
            group_mapping[src_id] = res.json()["id"]
        logger.info("Created Intent Groups.")

        raw_dashboard = json.dumps(dashboard)
        logger.info("Creating Dashboard.")
        for src_group, dst_group in group_mapping.items():
            raw_dashboard = raw_dashboard.replace(src_group, dst_group)

        res = self.ipf.put("settings/dashboard", json=json.loads(raw_dashboard))
        res.raise_for_status()
        logger.info("Created Dashboard.")
        return True



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Fhrp(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def group_state(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/group-state", sn=self.sn)

    @property
    def group_members(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/group-members", sn=self.sn)

    @property
    def stproot_alignment(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/stproot-alignment", sn=self.sn)

    @property
    def balancing(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/balancing", sn=self.sn)

    @property
    def glbp_forwarders(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/glbp-forwarders", sn=self.sn)

    @property
    def virtual_gateways(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/virtual-gateways", sn=self.sn)
